#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include <glib.h>
#include <glib-object.h>

#include <misc/utils.h>
#include <vdomain/vdomain.h>
#include <auth/auth.h>

#ifdef HAVE_CRYPT_H
#	include <crypt.h>
#endif

#include "md5.h"
#include "md5-crypt.h"

#define MD5_MARKER	"$1$"

#if !defined(HAVE_CRYPT_H) && defined(HAVE_CRYPT)
char *crypt(const char *key, const char *salt);
#endif

struct _VAuth {
	GObject		gobject;

	VDomainEntry	*entry;
};

struct _VAuthClass {
	GObjectClass	parent_class;
};

static void v_domain_auth_class_init(VDomainAuthClass *klass, gpointer data);

G_DEFINE_TYPE_WITH_CODE(VAuth, v_auth, G_TYPE_OBJECT,
                        G_IMPLEMENT_INTERFACE(V_TYPE_DOMAIN_AUTH,
                                              v_domain_auth_class_init));

static void v_auth_dispose(GObject *gobject);

static void v_auth_setup(VDomainAuth *auth, VDomainEntry *entry);
static gboolean v_auth_auth(VDomainAuth *auth, const gchar *pass,
                            GError **error);
static void v_auth_set(VDomainAuth *auth, const gchar *pass);

static void v_auth_class_init(VAuthClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->dispose = v_auth_dispose;
}

static void v_domain_auth_class_init(VDomainAuthClass *klass, gpointer data)
{
	klass->setup = v_auth_setup;
	klass->auth = v_auth_auth;
	klass->set = v_auth_set;
}

static void v_auth_init(VAuth *auth)
{
}

static void v_auth_dispose(GObject *gobject)
{
	VAuth *auth = V_AUTH(gobject);

	if (auth->entry) {
		g_object_unref(auth->entry);
		auth->entry = NULL;
	}
}

static gboolean v_auth_check(const gchar *pass, const gchar *crypted,
                             GError **error)
{
	const gchar *tmp;

	g_return_val_if_fail(pass, FALSE);
	g_return_val_if_fail(crypted, FALSE);
	g_return_val_if_fail(!error || !*error, FALSE);

	if (strncmp(crypted, MD5_MARKER, sizeof MD5_MARKER - 1) == 0)
		tmp = md5_crypt(pass, crypted);
	else {
#ifdef HAVE_CRYPT
		tmp = crypt(pass, crypted);
#else
		g_set_error(error, V_AUTH_ERROR, V_AUTH_ERROR_NOTSUPP,
		            "Authentication method not supported");
		return FALSE;
#endif
	}

	if (strcmp(tmp, crypted) != 0) {
		g_set_error(error, V_AUTH_ERROR, V_AUTH_ERROR_INVALID,
		            "Invalid or incorrect password");
		return FALSE;
	}

	return TRUE;
}

static G_CONST_RETURN gchar *v_auth_crypt(const gchar *pass)
{
	static const char passwd_chars[] =
	    "./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
	gchar salt[14] = MD5_MARKER;
	int i;

	g_return_val_if_fail(pass, NULL);

	for(i = 3; i < 11; i++)
		salt[i] = passwd_chars[g_random_int() % 64];

	return md5_crypt(pass, salt);
}

static void v_auth_setup(VDomainAuth *_auth, VDomainEntry *entry)
{
	VAuth *auth = V_AUTH(_auth);

	g_return_if_fail(V_IS_DOMAIN_ENTRY(entry));
	g_return_if_fail(!auth->entry);

	auth->entry = g_object_ref(entry);
}

static gboolean v_auth_auth(VDomainAuth *_auth, const gchar *pass,
                            GError **error)
{
	VAuth *auth = V_AUTH(_auth);

	g_return_val_if_fail(!error || !*error, FALSE);
	g_return_val_if_fail(V_IS_DOMAIN_ENTRY(auth->entry), FALSE);

	if (!pass) {
		g_set_error(error, V_AUTH_ERROR, V_AUTH_ERROR_INVALID,
		            "Password required");
		return FALSE;
	}

	if (!auth->entry->passwd) {
		g_set_error(error, V_AUTH_ERROR, V_AUTH_ERROR_DISABLED,
		            "Password disabled in this account");
		return FALSE;
	}

	return v_auth_check(pass, auth->entry->passwd, error);
}

static void v_auth_set(VDomainAuth *_auth, const gchar *pass)
{
	VAuth *auth = V_AUTH(_auth);

	g_return_if_fail(V_IS_DOMAIN_ENTRY(auth->entry));

	v_domain_entry_set_password(auth->entry,
	                            pass ? v_auth_crypt(pass) : NULL);
}
