#include <string.h>

#include <glib.h>
#include <gmodule.h>

#include <misc/utils.h>
#include <cdb/cdb.h>
#include "internal.h"

struct _CDBReader {
	GObject		gobject;

	GIOChannel	*io;
	guint32		eod;
	guint32		pos;
	gchar		header[CDB_HEADER_SIZE];
};

struct _CDBReaderClass {
	GObjectClass	parent_class;
};

G_DEFINE_TYPE(CDBReader, cdb_reader, G_TYPE_OBJECT);

static void cdb_reader_finalize(GObject *gobject);

static void cdb_reader_class_init(CDBReaderClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->finalize = cdb_reader_finalize;
}

static void cdb_reader_init(CDBReader *reader)
{
}

static void cdb_reader_finalize(GObject *gobject)
{
	CDBReader *reader = CDB_READER(gobject);

	if (reader->io) {
		g_io_channel_unref(reader->io);
		reader->io = NULL;
	}

	G_OBJECT_CLASS(cdb_reader_parent_class)->finalize(gobject);
}

CDBReader *cdb_reader_new(const gchar *filename, GError **error)
{
	GIOChannel *io;
	CDBReader *reader;

	g_return_val_if_fail(filename, NULL);
	g_return_val_if_fail(!error || !*error, NULL);

	io = g_io_channel_new_file(filename, "r", error);
	if (!io)
		return NULL;
	g_io_channel_set_encoding(io, NULL, NULL);

	reader = g_object_new(CDB_TYPE_READER, NULL);
	reader->io = io;

	if (!cdb_reader_firstrec(reader, error)) {
		g_object_unref(reader);
		reader = NULL;
	}

	return reader;
}

static void cdb_reader_abort(CDBReader *reader)
{
	if (reader->io) {
		g_io_channel_unref(reader->io);
		reader->io = NULL;
	}
}

gboolean cdb_reader_firstrec(CDBReader *reader, GError **error)
{
	GIOStatus status;
	gsize count;

	g_return_val_if_fail(CDB_IS_READER(reader), FALSE);
	g_return_val_if_fail(!error || !*error, FALSE);
	g_return_val_if_fail(reader->io, FALSE);

	status = g_io_channel_seek_position(reader->io, 0, G_SEEK_SET, error);
	if (status != G_IO_STATUS_NORMAL)
		goto out;

	status = g_io_channel_read_chars(reader->io, reader->header,
	                                 CDB_HEADER_SIZE, &count, error);
	if (status != G_IO_STATUS_NORMAL && status != G_IO_STATUS_EOF)
		goto out;
	if (count != CDB_HEADER_SIZE) {
		g_set_error(error, CDB_ERROR, CDB_ERROR_SHORT_READ,
		            "Short read while reading header.");
		goto out;
	}

	reader->pos = CDB_HEADER_SIZE;
	reader->eod = unpack(&reader->header[0]);

	return TRUE;

out:
	cdb_reader_abort(reader);
	return FALSE;
}

CDBDatum *cdb_reader_nextrec(CDBReader *reader, GError **error)
{
	GIOStatus status;
	guint32 klen, dlen;
	gchar buf[8];
	gchar *key, *data;
	gsize count;

	g_return_val_if_fail(CDB_IS_READER(reader), NULL);
	g_return_val_if_fail(!error || !*error, NULL);
	g_return_val_if_fail(reader->io, NULL);

	if (reader->pos >= reader->eod)
		return NULL;

	if (reader->eod - reader->pos < 8) {
		g_set_error(error, CDB_ERROR, CDB_ERROR_FORMAT,
		            "Incomplete record header.");
		goto out;
	}

	status = g_io_channel_seek_position(reader->io, reader->pos,
	                                    G_SEEK_SET, error);
	if (status != G_IO_STATUS_NORMAL)
		goto out;

	status = g_io_channel_read_chars(reader->io, buf, 8, &count, error);
	reader->pos += 8;
	if (status != G_IO_STATUS_NORMAL && status != G_IO_STATUS_EOF)
		goto out;
	if (count != 8) {
		g_set_error(error, CDB_ERROR, CDB_ERROR_SHORT_READ,
		            "Short read while reading record header.");
		goto out;
	}

	klen = unpack(&buf[0]);
	dlen = unpack(&buf[4]);

	if (reader->eod - reader->pos < klen) {
		g_set_error(error, CDB_ERROR, CDB_ERROR_FORMAT,
		            "Key too large for remaining space.");
		goto out;
	}
	reader->pos += klen;

	if (reader->eod - reader->pos < dlen) {
		g_set_error(error, CDB_ERROR, CDB_ERROR_FORMAT,
		            "Data too large for remaining space.");
		goto out;
	}
	reader->pos += dlen;

	key = g_malloc(klen + 1);
	status = g_io_channel_read_chars(reader->io, key, klen,
	                                 &count, error);
	if (status != G_IO_STATUS_NORMAL && status != G_IO_STATUS_EOF)
		goto out_key;
	if (count != klen) {
		g_set_error(error, CDB_ERROR, CDB_ERROR_SHORT_READ,
		            "Short key while reading record key.");
		goto out_key;
	}
	key[klen] = 0;

	data = g_malloc(dlen + 1);
	status = g_io_channel_read_chars(reader->io, data, dlen,
	                                 &count, error);
	if (status != G_IO_STATUS_NORMAL && status != G_IO_STATUS_EOF)
		goto out_data;
	if (count != dlen) {
		g_set_error(error, CDB_ERROR, CDB_ERROR_SHORT_READ,
		            "Short key while reading record data.");
		goto out_data;
	}
	data[dlen] = 0;

	return cdb_datum_new(klen, key, dlen, data, FALSE);

out_data:
	g_free(data);
out_key:
	g_free(key);
out:
	cdb_reader_abort(reader);
	return FALSE;
}

static gboolean match(GIOChannel *io, gsize klen,
                      const gchar *key, gboolean *matches, GError **error)
{
	gchar buf[32];

	*matches = TRUE;

	while(klen > 0) {
		GIOStatus status;
		gsize count, n = MIN(klen, sizeof buf);

		status = g_io_channel_read_chars(io, buf, n, &count, error);
		if (status != G_IO_STATUS_NORMAL && status != G_IO_STATUS_EOF)
			goto out;
		if (count != n) {
			g_set_error(error, CDB_ERROR, CDB_ERROR_SHORT_READ,
			            "Short key while reading record key.");
			goto out;
		}

		if (memcmp(buf, key, n) != 0) {
			*matches = FALSE;
			break;
		}

		key += n;
		klen -= n;
	}

	return TRUE;
out:
	return FALSE;
}

static gboolean cdb_reader_seek(CDBReader *reader, gsize klen,
                                const gchar *key, gsize *dlen, gboolean *found,
                                GError **error)
{
	guint32 hash = cdb_hash(klen, key);
	guint32 pos = (hash % CDB_HASH_ENTRIES) * 8;
	guint32 poshash = unpack(&reader->header[pos + 0]);
	guint32 lenhash = unpack(&reader->header[pos + 4]);
	guint32 h2;
	guint32 i;
	gsize count;

	*found = FALSE;

	if (!lenhash)
		return TRUE;

	h2 = (hash / CDB_HASH_ENTRIES) % lenhash;

	for(i = 0; i < lenhash; i++) {
		GIOStatus status;
		guint32 poskd;
		gchar buf[8];

		status = g_io_channel_seek_position(reader->io,
		         	poshash + (h2 * 8), G_SEEK_SET, error);
		if (status != G_IO_STATUS_NORMAL)
			goto out;

		status = g_io_channel_read_chars(reader->io, buf, 8,
		                                 &count, error);
		if (status != G_IO_STATUS_NORMAL && status != G_IO_STATUS_EOF)
			goto out;
		if (count != 8) {
			g_set_error(error, CDB_ERROR, CDB_ERROR_SHORT_READ,
			            "Short key while reading hash record.");
			goto out;
		}

		poskd = unpack(&buf[4]);
		if (!poskd)
			return TRUE;

		if (unpack(&buf[0]) == hash) {
			status = g_io_channel_seek_position(reader->io,
			         	poskd, G_SEEK_SET, error);
			if (status != G_IO_STATUS_NORMAL)
				goto out;

			reader->pos = poskd;

			status = g_io_channel_read_chars(reader->io, buf, 8,
			                                 &count, error);
			if (status != G_IO_STATUS_NORMAL &&
			    status != G_IO_STATUS_EOF)
				goto out;
			if (count != 8) {
				g_set_error(error, CDB_ERROR,
				            CDB_ERROR_SHORT_READ,
				            "Short key while reading record "
				            "header.");
				goto out;
			}

			if (unpack(&buf[0]) == klen) {
				if (!match(reader->io, klen, key, found,
				           error))
					goto out;
				if (*found) {
					*dlen = unpack(&buf[4]);
					return TRUE;
				}
			}
		}

		if (++h2 == lenhash)
			h2 = 0;
	}

	return TRUE;

out:
	cdb_reader_abort(reader);
	return FALSE;
}

CDBDatum *cdb_reader_get(CDBReader *reader, gsize klen, const gchar *key,
                         GError **error)
{
	GIOStatus status;
	gboolean found;
	gchar *data;
	gsize count, dlen = 0;

	g_return_val_if_fail(CDB_IS_READER(reader), NULL);
	g_return_val_if_fail(!error || !*error, NULL);
	g_return_val_if_fail(reader->io, NULL);

	if (!cdb_reader_seek(reader, klen, key, &dlen, &found, error)
	    || !found)
		return NULL;

	data = g_malloc(dlen + 1);
	status = g_io_channel_read_chars(reader->io, data, dlen,
	                                 &count, error);
	if (status != G_IO_STATUS_NORMAL && status != G_IO_STATUS_EOF)
		goto out_data;
	if (count != dlen) {
		g_set_error(error, CDB_ERROR, CDB_ERROR_SHORT_READ,
		            "Short key while reading record data.");
		goto out_data;
	}
	data[dlen] = 0;

	reader->pos += klen + dlen + 8;

	return cdb_datum_new(klen, g_strdup(key), dlen, data, FALSE);

out_data:
	g_free(data);
	return NULL;
}

CDBDatum *cdb_reader_get_string(CDBReader *reader, const gchar *key,
                                GError **error)
{
	return cdb_reader_get(reader, key ? strlen(key) : 0, key, error);
}
