#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>

#include <glib.h>
#include <glib/gstdio.h>
#include <gmodule.h>

#include <misc/utils.h>
#include <cdb/cdb.h>
#include "internal.h"

struct CDBHashEntry {
	guint32		hash;
	guint32		pos;
};

struct _CDBWriter {
	GObject		gobject;

	gchar		*filename;
	GIOChannel	*io;
	guint32		pos;
	GArray		*hashes[CDB_HASH_ENTRIES];
	gchar		header[CDB_HEADER_SIZE];
};

struct _CDBWriterClass {
	GObjectClass	parent_class;
};

G_DEFINE_TYPE(CDBWriter, cdb_writer, G_TYPE_OBJECT);

static void cdb_writer_finalize(GObject *gobject);

static void cdb_writer_class_init(CDBWriterClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->finalize = cdb_writer_finalize;
}

static void cdb_writer_init(CDBWriter *writer)
{
}

static void cdb_writer_finalize(GObject *gobject)
{
	CDBWriter *writer = CDB_WRITER(gobject);
	int i;

	if (writer->io) {
		g_io_channel_unref(writer->io);
		writer->io = NULL;
		g_unlink(writer->filename);
	}

	for(i = 0; i < CDB_HASH_ENTRIES; i++) {
		if (!writer->hashes[i])
			continue;

		g_array_free(writer->hashes[i], TRUE);
		writer->hashes[i] = NULL;
	}

	g_free(writer->filename);
	writer->filename = NULL;

	G_OBJECT_CLASS(cdb_writer_parent_class)->finalize(gobject);
}

static void cdb_writer_add(CDBWriter *writer, guint32 hash, guint32 pos)
{
	struct CDBHashEntry entry = { hash, pos };
	guint32 bucket = hash % CDB_HASH_ENTRIES;
	GArray *array = writer->hashes[bucket];

	if (!array) {
		array = g_array_new(FALSE, FALSE, sizeof entry);
		writer->hashes[bucket] = array;
	}
	g_array_append_val(array, entry);
}

static gboolean cdb_writer_write_hashes(CDBWriter *writer, GError **error)
{
	gboolean retval = FALSE;
	GArray *hashes = g_array_new(FALSE, FALSE, 8);
	int bucket;

	for(bucket = 0; bucket < CDB_HASH_ENTRIES; bucket++) {
		GIOStatus status;
		guint32 v, lenhash;
		GArray *array = writer->hashes[bucket];
		gsize count;
		int i;

		if (!array)
			lenhash = 0;
		else if (array->len <= 8)
			lenhash = array->len * 3 / 2;
		else
			lenhash = g_spaced_primes_closest(array->len * 3 / 2);

		pack(&writer->header[bucket * 8 + 0], writer->pos);
		pack(&writer->header[bucket * 8 + 4], lenhash);

		if (!lenhash)
			continue;

		g_array_set_size(hashes, lenhash);
		memset(hashes->data, 0, lenhash * 8);

		for(i = 0; i < array->len; i++) {
			struct CDBHashEntry *entry =
				&g_array_index(array, struct CDBHashEntry, i);
			guint32 subbucket =
				(entry->hash / CDB_HASH_ENTRIES) % lenhash;

			while(unpack(&hashes->data[subbucket * 8 + 4]))
				if (++subbucket >= lenhash)
					subbucket = 0;

			pack(&hashes->data[subbucket * 8 + 0], entry->hash);
			pack(&hashes->data[subbucket * 8 + 4], entry->pos);
		}

		status = g_io_channel_write_chars(writer->io, hashes->data,
		                                  lenhash * 8, &count, error);
		if (status != G_IO_STATUS_NORMAL)
			goto out;
		if (count != lenhash * 8) {
			g_set_error(error, CDB_ERROR, CDB_ERROR_SHORT_WRITE,
			            "Short write while writing hash record.");
			goto out;
		}

		pack(&writer->header[bucket * 8 + 0], writer->pos);
		pack(&writer->header[bucket * 8 + 4], lenhash);

		v = writer->pos + lenhash * 8;
		if (v < writer->pos) {
			g_set_error(error, CDB_ERROR, CDB_ERROR_FAILED,
			            "Resulting file too large.");
			goto out;
		}

		writer->pos = v;
	}

	retval = TRUE;
out:
	g_array_free(hashes, TRUE);
	return retval;
}

CDBWriter *cdb_writer_new(const gchar *filename, guint mode, GError **error)
{
	GIOChannel *io;
	GIOStatus status;
	CDBWriter *writer;
	gsize count;

	g_return_val_if_fail(filename, NULL);
	g_return_val_if_fail(!error || !*error, NULL);

	io = v_output_io_channel_new(filename, V_IO_FLAGS_CREATE |
	                                       V_IO_FLAGS_EXCL, mode, error);
	if (!io)
		return NULL;
	g_io_channel_set_encoding(io, NULL, NULL);

	writer = g_object_new(CDB_TYPE_WRITER, NULL);
	writer->filename = g_strdup(filename);
	writer->io = io;
	writer->pos = CDB_HEADER_SIZE;

	status = g_io_channel_write_chars(writer->io, writer->header,
	                                  CDB_HEADER_SIZE, &count, error);
	if (status != G_IO_STATUS_NORMAL) {
		g_object_unref(writer);
		return NULL;
	}
	if (count != CDB_HEADER_SIZE) {
		g_object_unref(writer);
		g_set_error(error, CDB_ERROR, CDB_ERROR_SHORT_WRITE,
		            "Short write while reading header.");
		return NULL;
	}

	return writer;
}

static void cdb_writer_abort(CDBWriter *writer)
{
	if (writer->io) {
		g_io_channel_unref(writer->io);
		writer->io = NULL;
		g_unlink(writer->filename);
	}
}

gboolean cdb_writer_putrec(CDBWriter *writer, CDBDatum *datum, GError **error)
{
	GIOStatus status;
	guint32 v;
	gchar buf[8];
	gsize count;

	g_return_val_if_fail(CDB_IS_WRITER(writer), FALSE);
	g_return_val_if_fail(datum, FALSE);
	g_return_val_if_fail(!error || !*error, FALSE);
	g_return_val_if_fail(writer->io, FALSE);

	pack(&buf[0], datum->klen);
	pack(&buf[4], datum->dlen);

	status = g_io_channel_write_chars(writer->io, buf, 8, &count, error);
	if (status != G_IO_STATUS_NORMAL)
		goto out;
	if (count != 8) {
		g_set_error(error, CDB_ERROR, CDB_ERROR_SHORT_WRITE,
		            "Short write while writing record header.");
		goto out;
	}

	status = g_io_channel_write_chars(writer->io, datum->key,
	                                  datum->klen, &count, error);
	if (status != G_IO_STATUS_NORMAL)
		goto out;
	if (count != datum->klen) {
		g_set_error(error, CDB_ERROR, CDB_ERROR_SHORT_WRITE,
		            "Short write while writing record key.");
		goto out;
	}

	status = g_io_channel_write_chars(writer->io, datum->data,
	                                  datum->dlen, &count, error);
	if (status != G_IO_STATUS_NORMAL)
		goto out;
	if (count != datum->dlen) {
		g_set_error(error, CDB_ERROR, CDB_ERROR_SHORT_WRITE,
		            "Short write while writing record date.");
		goto out;
	}

	cdb_writer_add(writer, cdb_hash(datum->klen, datum->key), writer->pos);

	v = writer->pos + 8 + datum->klen + datum->dlen;
	if (v < writer->pos) {
		g_set_error(error, CDB_ERROR, CDB_ERROR_FAILED,
		            "Resulting file too large.");
		goto out;
	}
	writer->pos = v;

	return TRUE;

out:
	cdb_writer_abort(writer);
	return FALSE;
}

gboolean cdb_writer_finish(CDBWriter *writer, const gchar *filename,
                           GError **error)
{
	GIOStatus status;
	gsize count;

	g_return_val_if_fail(CDB_IS_WRITER(writer), FALSE);
	g_return_val_if_fail(filename, FALSE);
	g_return_val_if_fail(!error || !*error, FALSE);
	g_return_val_if_fail(writer->io, FALSE);

	if (!cdb_writer_write_hashes(writer, error))
		goto out;

	status = g_io_channel_seek_position(writer->io, 0, G_SEEK_SET, error);
	if (status != G_IO_STATUS_NORMAL)
		goto out;

	status = g_io_channel_write_chars(writer->io, writer->header,
	                                  CDB_HEADER_SIZE, &count, error);
	if (status != G_IO_STATUS_NORMAL)
		goto out;
	if (count != CDB_HEADER_SIZE) {
		g_set_error(error, CDB_ERROR, CDB_ERROR_SHORT_WRITE,
		            "Short write while writing header.");
		goto out;
	}

	status = g_io_channel_flush(writer->io, error);
	if (status != G_IO_STATUS_NORMAL)
		goto out;

	g_io_channel_unref(writer->io);
	writer->io = NULL;

	if (g_rename(writer->filename, filename) < 0) {
		g_set_error(error, G_FILE_ERROR,
		            g_file_error_from_errno(errno),
		            g_strerror(errno));
		goto out;
	}

	return TRUE;

out:
	cdb_writer_abort(writer);
	return FALSE;
}
