#define _GNU_SOURCE
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <locale.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <glib.h>
#include <glib/gstdio.h>

#define G_IMPLEMENT_INLINES 1
#define __MISC_UTILS_C__
#include <misc/utils.h>
#undef G_IMPLEMENT_INLINES

#include <misc/vconfig.h>

#define NAME_TO_PATH(name)	(#name)

#define DEFAULT_SCRIPT_DIR	NAME_TO_PATH(PACKAGE)

static void log_handler(const gchar *domain, GLogLevelFlags level,
                        const gchar *message, gpointer user_data)
{
	const gchar *msg;

	if (level & G_LOG_LEVEL_ERROR)
		msg = N_("Error");
	else if (level & G_LOG_LEVEL_CRITICAL)
		msg = N_("Critical");
	else if (level & G_LOG_LEVEL_WARNING)
		msg = N_("Warning");
	else if (level & G_LOG_LEVEL_MESSAGE)
		msg = N_("Message");
	else if (level & G_LOG_LEVEL_INFO)
		msg = N_("Info");
	else
		msg = N_("Debug");

	fprintf(stderr, "%s: %s\n", _(msg), message);
}

void v_init(gint *argc, gchar ***argv, const GOptionEntry *entries,
            void (*opt_setup)(GOptionContext *), const gchar *args,
            const gchar *description)
{
	static GOptionContext *context = NULL;

	static gboolean initialized = FALSE;

	if (initialized) {
		if (argc || argv || entries || opt_setup ||
		    args || description)
			g_warning(_("CLI initialization called twice."));
		return;
	}
	initialized = TRUE;

	g_type_init();

	g_log_set_always_fatal(G_LOG_FATAL_MASK | G_LOG_LEVEL_CRITICAL);

	g_log_set_default_handler(log_handler, NULL);

	if (argc && argv) {
		GError *error = NULL;
		gchar *tmp;

		g_return_if_fail(!context);

		setlocale(LC_ALL, "");
		bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
#ifdef HAVE_BIND_TEXT_DOMAIN_CODESET
		bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
#endif
		textdomain(GETTEXT_PACKAGE);

		if (args || description)
			tmp = g_strconcat(args ? args : "",
			                  (args && description) ? "\n\n" : "",
			                  description ? description : "",
			                  NULL);
		else
			tmp = NULL;
		context = g_option_context_new(tmp);
		g_free(tmp);

		g_option_context_set_ignore_unknown_options(context, FALSE);

		if (entries)
			g_option_context_add_main_entries(context, entries,
			                                  NULL);

		if (opt_setup)
			opt_setup(context);

		if (!g_option_context_parse(context, argc, argv, &error)) {
			v_init_error(error ? error->message
			                   : _("Option parse error"));
			g_error_free(error);
		}

		g_option_context_free(context);
	} else if (argc || argv || entries || opt_setup ||
	           args || description)
		g_warning(_("CLI initialization call incomplete."));
}

void v_init_error(const gchar *msg)
{
	g_print("%s: %s\n", g_get_prgname(), msg);
	g_print(_("Try `%s --help' for more information.\n"),
	        g_get_prgname());
	exit(1);
}

G_CONST_RETURN gchar *v_get_hostname(void)
{
	static gchar *hostname = NULL;
	gsize len = 512;

	if (hostname)
		return hostname;

	hostname = g_strdup(g_getenv("TCPLOCALHOST"));
	if (hostname)
		return hostname;

	hostname = g_malloc(len);
retry:
	if (gethostname(hostname, len - 1) < 0) {
		if (errno == ENAMETOOLONG) {
			hostname = g_realloc(hostname, len <<= 1);
			goto retry;
		}

		hostname = NULL;
	}

	if (hostname && strlen(hostname) + 1 < len)
		hostname = g_realloc(hostname, strlen(hostname) + 1);

	return hostname;
}

GIOChannel *v_output_io_channel_new(const gchar *filename, VIOFlags vflags,
                                    guint mode, GError **error)
{
	GIOChannel *io;
	GIOFlags ioflags;
	int flags = O_WRONLY;
	int fd;

	g_return_val_if_fail(filename, NULL);
	g_return_val_if_fail(!error || !*error, NULL);

	if (vflags & V_IO_FLAGS_CREATE)
		flags |= O_CREAT;
	if (vflags & V_IO_FLAGS_EXCL)
		flags |= O_EXCL;
	if (vflags & V_IO_FLAGS_TRUNC)
		flags |= O_TRUNC;
	if (vflags & V_IO_FLAGS_APPEND)
		flags |= O_APPEND;

	fd = g_open(filename, flags, mode);
	if (fd < 0) {
		g_set_error(error, G_FILE_ERROR,
		            g_file_error_from_errno(errno),
		            g_strerror(errno));
		return NULL;
	}

	io = g_io_channel_unix_new(fd);
	if (!io)
		return NULL;

	g_io_channel_set_close_on_unref(io, TRUE);

	ioflags = g_io_channel_get_flags(io);
	if ((ioflags & (G_IO_FLAG_IS_SEEKABLE | G_IO_FLAG_IS_WRITEABLE))
	            != (G_IO_FLAG_IS_SEEKABLE | G_IO_FLAG_IS_WRITEABLE)) {
		g_io_channel_unref(io);
		return NULL;
	}

	return io;
}

GIOStatus v_io_channel_sync(GIOChannel *io, GError **error)
{
	GIOStatus status;
	int fd;

	g_return_val_if_fail(io, G_IO_STATUS_ERROR);
	g_return_val_if_fail(!error || !*error, G_IO_STATUS_ERROR);

	status = g_io_channel_flush(io, error);
	if (status != G_IO_STATUS_NORMAL)
		return status;

	fd = g_io_channel_unix_get_fd(io);
	if (fsync(fd) < 0) {
		g_set_error(error, G_FILE_ERROR,
		            g_file_error_from_errno(errno),
		            g_strerror(errno));
		return G_IO_STATUS_ERROR;
	}

	return G_IO_STATUS_NORMAL;
}

gboolean v_drop_process_rights(const gchar *username, uid_t uid, gid_t gid,
                               GError **error)
{
	gchar *msg = NULL;
	uid_t old_uid = getuid();
	gid_t old_gid = getgid();

	g_return_val_if_fail(!error || !*error, FALSE);

	if (setresgid(gid, gid, gid) < 0) {
		msg = g_strdup_printf(_("Could not drop gid privileges "
		                         "to group `%d': %%s"), (int)gid);
		goto out_errno;
	}

	if (setresuid(uid, uid, uid) < 0) {
		msg = g_strdup_printf(_("Could not drop uid privileges "
		                        "to user `%d': %%s"), (int)uid);
		goto out_errno;
	}

	if ((old_gid != gid && (setgid(old_gid) != -1
	                        || setegid(old_gid) != -1))
	    || getgid() != gid || getegid() != gid) {
		msg = g_strdup_printf(_("Could not drop gid privileges to "
		                        "group `%d'"), (int)gid);
		goto out;
	}

	if ((old_uid != uid && (setuid(old_uid) != -1
	                        || seteuid(old_uid) != -1))
	    || getuid() != uid || geteuid() != uid) {
		msg = g_strdup_printf(_("Could not drop uid privileges to "
		                        "user `%d'"), (int)uid);
		goto out;
	}

	return TRUE;

out_errno:
	if (error)
		g_set_error(error, G_FILE_ERROR,
		            g_file_error_from_errno(errno),
		            msg, g_strerror(errno));
out:
	if (error && !*error)
		g_set_error(error, G_FILE_ERROR, G_FILE_ERROR_FAILED, msg);
	g_free(msg);
	return FALSE;
}

gint v_exec(gchar **argv, GError **error)
{
	gint retval = 127;

	g_return_val_if_fail(!error || !*error, -1);

	if (!g_spawn_sync(NULL, argv, NULL, G_SPAWN_LEAVE_DESCRIPTORS_OPEN ||
	                                    G_SPAWN_CHILD_INHERITS_STDIN,
	                  NULL, NULL, NULL, NULL, &retval, error))
		goto out;

	if (!WIFEXITED(retval))
		retval = -1;
	else
		retval = WEXITSTATUS(retval);
	if (retval == 99)
		retval = 0;
out:

	return retval;
}

gint v_exec_script(const gchar *script_name)
{
	gchar *path, *script = NULL;
	gchar *argv[2];
	gint retval = 0;

	path = v_config_get_string(NULL, NULL, "script-directory");
	if (!path)
		path = DEFAULT_SCRIPT_DIR;
	if (path[0] == '/')
		script = g_strconcat(path, "/", script_name, NULL);
	else
		script = g_strconcat(SYSCONFDIR "/", path, "/",
		                     script_name, NULL);
	if (path != DEFAULT_SCRIPT_DIR)
		g_free(path);

	if (!g_file_test(script, G_FILE_TEST_IS_REGULAR |
	                         G_FILE_TEST_IS_EXECUTABLE))
		goto out;

	argv[0] = script;
	argv[1] = NULL;

	retval = v_exec(argv, NULL);

out:
	g_free(script);
	return retval;
}

gchar *v_get_password(void)
{
	if (isatty(0)) {
		gchar *first, *second;

		first = g_strdup(getpass(_("Enter the password: ")));
		if (!first)
			return NULL;

		second = g_strdup(getpass(_("Type again for verification: ")));
		if (!second || strcmp(first, second) != 0) {
			g_free(first);
			g_free(second);
			fprintf(stderr, _("%s: Passwords don't match.\n"),
			        g_get_prgname());
			return NULL;
		}
		g_free(second);

		if (!first[0]) {
			fprintf(stderr, _("%s: Empty password.\n"),
			        g_get_prgname());
			g_free(first);
			return NULL;
		}

		return first;
	} else {
		GString *pass = g_string_sized_new(128);

		for(;;) {
			int c = getchar();

			if (c <= 0 || c == EOF || c == '\n' || c == '\r')
				break;

			g_string_append_c(pass, c);
		}

		return g_string_free(pass, FALSE);
	}
}

gboolean v_mkdir_recursive(const gchar *path, guint mode,
                           guint fill_mode, GError **error)
{
	gboolean retval;
	gchar *tmp, *p;
	gsize len;

	g_return_val_if_fail(path, FALSE);
	g_return_val_if_fail(!error || !*error, FALSE);

	len = strlen(path);
	tmp = g_malloc(len + 1);
	memcpy(tmp, path, len + 1);

	for(p = &tmp[len]; p; p = strrchr(tmp, '/')) {
		*p = 0;
		if (g_file_test(tmp, G_FILE_TEST_EXISTS))
			break;
	}

	while(p < &tmp[len]) {
		*p = '/';
		p += strlen(p);
		if (g_mkdir(tmp, p < &tmp[len] ? fill_mode : mode) < 0) {
			g_set_error(error, G_FILE_ERROR,
			            g_file_error_from_errno(errno),
			            g_strerror(errno));
			break;
		}
	}
	retval = (p == &tmp[len]);

	g_free(tmp);
	return retval;
}
