#include <string.h>
#include <stdlib.h>

#include <glib.h>
#include <gmodule.h>

#include <misc/utils.h>
#include <misc/vconfig.h>

struct _VConfig {
	GObject		gobject;

	GKeyFile	*keys;
};

struct _VConfigClass {
	GObjectClass	parent_class;
};

G_DEFINE_TYPE(VConfig, v_config, G_TYPE_OBJECT);

G_LOCK_DEFINE_STATIC(_default_lock);
static VConfig *_default_config = NULL;

static void v_config_finalize(GObject *gobject);

static void v_config_class_init(VConfigClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->finalize = v_config_finalize;
}

static void v_config_init(VConfig *config)
{
}

static void v_config_finalize(GObject *gobject)
{
	VConfig *config = V_CONFIG(gobject);

	g_key_file_free(config->keys);
	config->keys = NULL;

	G_OBJECT_CLASS(v_config_parent_class)->finalize(gobject);
}

VConfig *v_config_new(const gchar *filename, GError **error)
{
	GKeyFile *file = g_key_file_new();
	VConfig *config;

	g_return_val_if_fail(!error || !*error, NULL);

	if (!filename)
		filename = SYSCONFDIR "/" PACKAGE ".conf";

	if (!g_key_file_load_from_file(file, filename,
	                               G_KEY_FILE_NONE, error)) {
		g_key_file_free(file);
		return NULL;
	}

	config = g_object_new(V_TYPE_CONFIG, NULL);
	config->keys = file;

	return config;
}

VConfig *v_config_get_default(void)
{
	GError *error = NULL;
	VConfig *config = NULL;

	G_LOCK(_default_lock);
	if (!_default_config)
		_default_config = v_config_new(NULL, &error);
	if (_default_config)
		config = g_object_ref(_default_config);
	else {
		g_warning(_("Could not open default config: %s\n"),
		          error ? error->message : _("Unknown"));
		g_error_free(error);
	}
	G_UNLOCK(_default_lock);

	return config;
}

void v_config_invalidate_default(void)
{
	G_LOCK(_default_lock);
	if (_default_config)
		g_object_unref(_default_config);
	_default_config = NULL;
	G_UNLOCK(_default_lock);
}

typedef gpointer (*VConfigGetFunc)(GKeyFile *keys, const gchar *group,
                                   const gchar *key, GError **error);

static gpointer v_config_string_getter(GKeyFile *keys, const gchar *group,
                                       const gchar *key, GError **error)
{
	return g_key_file_get_string(keys, group, key, error);
}

static gpointer v_config_strv_getter(GKeyFile *keys, const gchar *group,
                                     const gchar *key, GError **error)
{
	return g_key_file_get_string_list(keys, group, key, NULL, error);
}

static gpointer v_config_int_getter(GKeyFile *keys, const gchar *group,
                                    const gchar *key, GError **error)
{
	return GINT_TO_POINTER(g_key_file_get_integer(keys, group,
	                                              key, error));
}

static gboolean v_config_get(VConfig *config, const char *group,
                             const gchar *key, gpointer *result,
                             VConfigGetFunc func)
{
	gboolean release_config = FALSE, retval = FALSE;
	GError *error = NULL;

	g_return_val_if_fail(!config || V_IS_CONFIG(config), FALSE);
	g_return_val_if_fail(key, FALSE);

	if (!group)
		group = "general";

	if (!config) {
		config = v_config_get_default();
		if (!config)
			return FALSE;
		release_config = TRUE;
	}

	*result = func(config->keys, group, key, &error);
	if (error) {
		if (error->domain != G_KEY_FILE_ERROR ||
                    (error->code != G_KEY_FILE_ERROR_NOT_FOUND &&
		     error->code != G_KEY_FILE_ERROR_KEY_NOT_FOUND &&
		     error->code != G_KEY_FILE_ERROR_GROUP_NOT_FOUND))
			g_warning(_("Error reading key `%s' in group `%s': "
			           " %s"), group, key, error->message);
		g_error_free(error);
	} else
		retval = TRUE;

	if (release_config);
		g_object_unref(config);

	return retval;
}

gchar *v_config_get_string(VConfig *config, const gchar *group,
                           const gchar *key)
{
	gpointer result;

	if (!v_config_get(config, group, key, &result,
	                  v_config_string_getter))
		g_free(result);

	return (gchar *)result;
}

gchar **v_config_get_strv(VConfig *config, const gchar *group,
                          const gchar *key)
{
	gpointer result;

	if (!v_config_get(config, group, key, &result,
	                  v_config_strv_getter))
		g_free(result);

	return (gchar **)result;
}

int v_config_get_int(VConfig *config, const gchar *group, const gchar *key)
{
	gpointer result;

	if (!v_config_get(config, group, key, &result,
	                  v_config_int_getter))
		return -1;

	return GPOINTER_TO_INT(result);
}

gboolean v_config_get_ulong(VConfig *config, const gchar *group,
                            const gchar *key, gulong *retval)
{
	gulong val;
	gchar *tmp, *end;

	g_return_val_if_fail(retval, FALSE);

	tmp = v_config_get_string(config, group, key);
	if (!tmp)
		return FALSE;
	val = strtoul(tmp, &end, 0);
	g_free(tmp);

	if (end && *end)
		return FALSE;

	*retval = val;
	return TRUE;
}

gboolean v_config_get_uint(VConfig *config, const gchar *group,
                           const gchar *key, guint *retval)
{
	gulong val;

	g_return_val_if_fail(retval, FALSE);

	if (!v_config_get_ulong(config, group, key, &val))
		return FALSE;

	if (val > G_MAXUINT)
		return FALSE;

	*retval = (guint)val;
	return TRUE;
}
