#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <errno.h>

#include <glib.h>
#include <glib/gstdio.h>
#include <glib-object.h>

#include <misc/utils.h>
#include <qmail/qmail.h>

#define BUFFER_SIZE		4096

static const gchar *_dirs[] = { "/cur", "/new", "/tmp" };

#define V_MAILDIR_COUNT	G_N_ELEMENTS(_dirs)

struct _VMaildir {
	GObject			gobject;

	gchar			*base;
	gchar			*dirs[V_MAILDIR_COUNT];
	int			valid;
	GSList			*pending;
};

struct _VMaildirClass {
	GObjectClass		parent_class;
};

G_DEFINE_TYPE(VMaildir, v_maildir, G_TYPE_OBJECT);

static void v_maildir_dispose(GObject *gobject);
static void v_maildir_finalize(GObject *gobject);

static void v_maildir_class_init(VMaildirClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->dispose = v_maildir_dispose;
	gobject_class->finalize = v_maildir_finalize;
}

static void v_maildir_init(VMaildir *maildir)
{
	maildir->valid = TRUE;
}

static void v_maildir_dispose(GObject *gobject)
{
	VMaildir *maildir = V_MAILDIR(gobject);

	while(maildir->pending) {
		GSList *next = maildir->pending->next;
		gchar *tmp, *filename = maildir->pending->data;

		g_slist_free_1(maildir->pending);
		maildir->pending = next;

		tmp = g_strconcat(maildir->dirs[V_MAILDIR_TMP], "/",
		                  filename, NULL);
		g_unlink(tmp);
		g_free(tmp);

		g_free(filename);
	}
}

static void v_maildir_finalize(GObject *gobject)
{
	VMaildir *maildir = V_MAILDIR(gobject);
	int i;

	g_free(maildir->base);
	for(i = 0; i < V_MAILDIR_COUNT; i++)
		g_free(maildir->dirs[i]);

	G_OBJECT_CLASS(v_maildir_parent_class)->finalize(gobject);
}

static VMaildir *v_maildir_new_internal(const gchar *base)
{
	VMaildir *maildir = g_object_new(V_TYPE_MAILDIR, NULL);
	int i;

	maildir->base = g_strdup(base);
	for(i = 0; i < V_MAILDIR_COUNT; i++)
		maildir->dirs[i] = g_strconcat(base, _dirs[i], NULL);

	return maildir;
}

VMaildir *v_maildir_new(const gchar *name, GError **error)
{
	VMaildir *maildir;
	int i;

	g_return_val_if_fail(name, NULL);
	g_return_val_if_fail(!error || !*error, NULL);

	maildir = v_maildir_new_internal(name);
	for(i = 0; i < V_MAILDIR_COUNT; i++)
		if (!g_file_test(maildir->dirs[i], G_FILE_TEST_IS_DIR))
			break;

	if (i < V_MAILDIR_COUNT) {
		g_set_error(error, V_QMAIL_ERROR, V_QMAIL_ERROR_INVALID,
		            "Invalid maildir, missing directory `%s'.",
		            maildir->dirs[i]);

		g_object_unref(maildir);
		maildir = NULL;
	}

	return maildir;
}

VMaildir *v_maildir_new_create(const gchar *name, GError **error)
{
	VMaildir *maildir;
	int i;

	g_return_val_if_fail(!error || !*error, NULL);

	maildir = v_maildir_new_internal(name);

	if (!v_mkdir_recursive(maildir->base, 0700, 0755, error))
		goto out;

	for(i = 0; i < V_MAILDIR_COUNT; i++)
		if (g_mkdir(maildir->dirs[i], 0755))
			break;

	if (i < V_MAILDIR_COUNT) {
		g_set_error(error, G_FILE_ERROR,
		            g_file_error_from_errno(errno),
		            g_strerror(errno));
		for(; i >= 0; i--)
			g_rmdir(maildir->dirs[i]);
		goto out;
	}

	return maildir;

out:
	g_object_unref(maildir);
	return NULL;
}

gchar *v_maildir_get_base(VMaildir *maildir)
{
	g_return_val_if_fail(V_IS_MAILDIR(maildir), FALSE);
	g_return_val_if_fail(maildir->valid, FALSE);

	return g_strdup(maildir->base);
}

static gboolean delete_directory(GString *entry, GError **error)
{
	const gchar *name;
	GDir *dir = g_dir_open(entry->str, 0, error);
	int len;
	int retval = FALSE;

	if (!dir)
		return FALSE;

	g_string_append_c(entry, '/');
	len = strlen(entry->str);

	while((name = g_dir_read_name(dir))) {
		g_string_truncate(entry, len);
		g_string_append(entry, name);

		if (g_file_test(entry->str, G_FILE_TEST_IS_DIR)) {
			if (!delete_directory(entry, error))
				goto out;
			if (g_rmdir(entry->str) < 0)
				goto out;
		} else {
			if (g_unlink(entry->str) < 0)
				goto out;
		}
	}

	retval = TRUE;
out:
	if (!retval && error && !*error)
		g_set_error(error, G_FILE_ERROR,
		            g_file_error_from_errno(errno),
		            g_strerror(errno));
	g_dir_close(dir);

	return retval;
}

gboolean v_maildir_delete(VMaildir *maildir, GError **error)
{
	GString *entry;
	int retval;

	g_return_val_if_fail(V_IS_MAILDIR(maildir), FALSE);
	g_return_val_if_fail(!error || !*error, FALSE);
	g_return_val_if_fail(maildir->valid, FALSE);

	maildir->valid = FALSE;

	entry = g_string_new(maildir->base);
	retval = delete_directory(entry, error);
	g_string_free(entry, TRUE);

	if (retval && g_rmdir(maildir->base) < 0) {
		g_set_error(error, G_FILE_ERROR,
		            g_file_error_from_errno(errno),
		            g_strerror(errno));
		retval = FALSE;
	}

	return retval;
}

gpointer v_maildir_prepare(VMaildir *maildir, GIOChannel *in, gchar **headers,
                           GError **error)
{
	GIOChannel *out = NULL;
	GIOStatus status;
	GError *my_error = NULL;
	gpointer handle = NULL;
	gchar *tmp, *buf = NULL, *filename = NULL;
	gsize size, count;

	g_return_val_if_fail(V_IS_MAILDIR(maildir), FALSE);
	g_return_val_if_fail(in, FALSE);
	g_return_val_if_fail(!error || !*error, FALSE);

	if (g_io_channel_get_encoding(in)) {
		status = g_io_channel_set_encoding(in, NULL, error);
		if (status != G_IO_STATUS_NORMAL)
			return NULL;
	}

	if (!g_file_test(maildir->dirs[V_MAILDIR_TMP], G_FILE_TEST_IS_DIR) ||
	    !g_file_test(maildir->dirs[V_MAILDIR_NEW], G_FILE_TEST_IS_DIR)) {
		g_set_error(error, V_QMAIL_ERROR, V_QMAIL_ERROR_INVALID,
		            _("Destination maildir `%s' broken"),
		            maildir->base);
		goto out;
	}

retry:
	filename = g_strdup_printf("%ld.%ld.%s", (long)time(NULL),
	                           (long)getpid(), v_get_hostname());

	tmp = g_strconcat(maildir->dirs[V_MAILDIR_TMP], "/", filename, NULL);
	out = v_output_io_channel_new(tmp, V_IO_FLAGS_EXCL |
	                                   V_IO_FLAGS_CREATE, 0600, &my_error);
	g_free(tmp);
	if (!out) {
		if (my_error->domain == G_FILE_ERROR &&
		    my_error->code == G_FILE_ERROR_EXIST) {
			g_error_free(my_error);
			my_error = NULL;
			g_usleep(100000);
			g_free(filename);
			goto retry;
		}
		g_propagate_error(error, my_error);
		goto out;
	}
	g_io_channel_set_encoding(out, NULL, NULL);

	for(; headers && *headers; headers++) {
		size = strlen(*headers);
		status = g_io_channel_write_chars(out, *headers, size,
		                                  &count, error);
		if (status != G_IO_STATUS_NORMAL)
			goto out;
		if (count != size) {
			g_set_error(error, V_QMAIL_ERROR,
			            V_QMAIL_ERROR_SHORT_WRITE,
			            "Short write while writing mail header.");
			goto out;
		}
	}

	buf = g_malloc(BUFFER_SIZE);
	for(;;) {
		status = g_io_channel_read_chars(in, buf, BUFFER_SIZE,
		                                 &size, error);
		if (status != G_IO_STATUS_NORMAL &&
		    status != G_IO_STATUS_EOF)
			goto out;

		if (!size)
			break;

		status = g_io_channel_write_chars(out, buf, size,
		                                  &count, error);
		if (status != G_IO_STATUS_NORMAL)
			goto out;
		if (count != size) {
			g_set_error(error, V_QMAIL_ERROR,
			            V_QMAIL_ERROR_SHORT_WRITE,
			            "Short write while writing mail content.");
			goto out;
		}
	}

	status = v_io_channel_sync(out, error);
	if (status != G_IO_STATUS_NORMAL)
		goto out;

	handle = filename;
	filename = NULL;
	maildir->pending = g_slist_prepend(maildir->pending, handle);

out:
	if (out) {
		g_io_channel_unref(out);
		if (!handle) {
			tmp = g_strconcat(maildir->dirs[V_MAILDIR_TMP], "/",
			                  filename, NULL);
			g_unlink(tmp);
			g_free(tmp);
		}
	}

	g_free(buf);
	g_free(filename);

	return handle;
}

gboolean v_maildir_deliver(VMaildir *maildir, gpointer handle, GError **error)
{
	gboolean retval = FALSE;
	GSList *elem;
	gchar *src, *dst, *p, *filename = handle;
	int i;

	g_return_val_if_fail(V_IS_MAILDIR(maildir), FALSE);
	g_return_val_if_fail(handle, FALSE);
	g_return_val_if_fail(!error || !*error, FALSE);

	elem = g_slist_find(maildir->pending, handle);
	g_return_val_if_fail(elem, FALSE);

	p = strchr(filename, '.');
	g_return_val_if_fail(p, FALSE);
	p = strchr(p + 1, '.');
	g_return_val_if_fail(p, FALSE);

	src = g_strconcat(maildir->dirs[V_MAILDIR_TMP], "/",
	                  filename, NULL);
	dst = g_strconcat(maildir->dirs[V_MAILDIR_NEW], "/",
	                  filename, NULL);
	*p = 0;
	for(i = 1;; i++) {
		char buf[10];

		if (link(src, dst) == 0) {
			g_unlink(src);
			maildir->pending =
				g_slist_delete_link(maildir->pending, elem);
			break;
		}

		if (errno != EEXIST) {
			g_set_error(error, G_FILE_ERROR,
			            g_file_error_from_errno(errno),
			            g_strerror(errno));
			goto out;
		}

		if (i >= 65536) {
			g_set_error(error, V_QMAIL_ERROR,
			            V_QMAIL_ERROR_FAILED,
			            _("Could not find unique filename."));
			goto out;
		}

		g_free(dst);
		snprintf(buf, 8, "_%d.", i);
		dst = g_strconcat(maildir->dirs[V_MAILDIR_NEW], "/",
		                  filename, buf, p + 1, NULL);
	}

	retval = TRUE;
out:
	*p = '.';
	g_free(src);
	g_free(dst);
	g_free(filename);

	return retval;
}

void v_maildir_abort(VMaildir *maildir, gpointer handle)
{
	GSList *elem;
	gchar *tmp, *filename = handle;

	g_return_if_fail(V_IS_MAILDIR(maildir));
	g_return_if_fail(handle);

	elem = g_slist_find(maildir->pending, handle);
	g_return_if_fail(elem);
	maildir->pending = g_slist_delete_link(maildir->pending, elem);

	tmp = g_strconcat(maildir->dirs[V_MAILDIR_TMP], "/",
	                  filename, NULL);
	g_unlink(tmp);
	g_free(tmp);

	g_free(filename);
}

