#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>  
#include <sys/stat.h>

#include <glib.h>
#include <glib/gstdio.h>
#include <glib-object.h>

#include <misc/utils.h>
#include <misc/vconfig.h>
#include <qmail/qmail.h>

#define LOCAL_DOMAINS_FILE	"/control/locals"
#define VIRTUAL_DOMAINS_FILE	"/control/virtualdomains"

struct _VQMailVirtual {
	GObject			gobject;

	gchar			*filename_local;
	gchar			*filename_virtual;
	GHashTable		*entries;
	time_t			ctime_local;
	time_t			ctime_virtual;
};

struct _VQMailVirtualClass {
	GObjectClass		parent_class;
};

G_DEFINE_TYPE(VQMailVirtual, v_qmail_virtual, G_TYPE_OBJECT);

G_LOCK_DEFINE_STATIC(_default_lock);
static VQMailVirtual *_default_virtual = NULL;

static void v_qmail_virtual_finalize(GObject *gobject);

static void v_qmail_virtual_class_init(VQMailVirtualClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->finalize = v_qmail_virtual_finalize;
}

static void v_qmail_virtual_init(VQMailVirtual *qmail_virtual)
{
	qmail_virtual->entries = 
		g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
}

static void v_qmail_virtual_finalize(GObject *gobject)
{
	VQMailVirtual *qmail_virtual = V_QMAIL_VIRTUAL(gobject);

	g_free(qmail_virtual->filename_local);
	qmail_virtual->filename_local = NULL;

	g_free(qmail_virtual->filename_virtual);
	qmail_virtual->filename_virtual = NULL;

	g_hash_table_destroy(qmail_virtual->entries);
	qmail_virtual->entries = NULL;

	G_OBJECT_CLASS(v_qmail_virtual_parent_class)->finalize(gobject);
}

static gboolean lookup_virtual_domain(GHashTable *entries,
                                      const gchar *domain,
                                      gpointer *result)
{
	gpointer key;

	if (g_hash_table_lookup_extended(entries, domain, &key, result))
		return TRUE;
	if (!domain[0])
		return FALSE;

	while((domain = strchr(domain + 1, '.')) != NULL)
		if (g_hash_table_lookup_extended(entries, domain,
		                                 &key, result))
			return TRUE;

	return g_hash_table_lookup_extended(entries, "", &key, result);
}

static void add_virtual_domain(GHashTable *entries, gchar *domain, gchar *user)
{
	gpointer result;

	if (lookup_virtual_domain(entries, domain, &result)) {
		g_free(domain);
		g_free(user);
		return;
	}

	g_hash_table_insert(entries, domain, user ? g_strstrip(user) : NULL);
}

G_CONST_RETURN gchar *v_qmail_virtual_lookup(VQMailVirtual *virtual,
                                             const gchar *domain)
{
	gpointer result = NULL;

	g_return_val_if_fail(V_IS_QMAIL_VIRTUAL(virtual), NULL);
	g_return_val_if_fail(domain, NULL);

	lookup_virtual_domain(virtual->entries, domain, &result);
	return (const gchar *)result;
}

static gboolean parse_file(GHashTable *entries, const gchar *filename,
                           int n_tokens, GError **error)
{
	GIOChannel *io;
	GIOStatus status;
	GString *line;

	g_assert(n_tokens == 1 || n_tokens == 2);

	io = g_io_channel_new_file(filename, "r", error);
	if (!io)
		return FALSE;

	status = g_io_channel_set_encoding(io, NULL, error);
	if (status != G_IO_STATUS_NORMAL)
		return FALSE;

	line = g_string_new(NULL);
	for(;;) {
		gchar **tokens;
		gsize end;

		status = g_io_channel_read_line_string(io, line, &end, error);
		if (status == G_IO_STATUS_EOF)
			break;
		if (status != G_IO_STATUS_NORMAL)
			goto out;
		g_string_truncate(line, end);

		tokens = g_strsplit(line->str, ":", n_tokens);
		if (!tokens || !tokens[0] || (n_tokens > 1 && !tokens[1])) {
			g_strfreev(tokens);
			continue;
		}

		add_virtual_domain(entries, tokens[0], tokens[1]);
	}
	g_string_free(line, TRUE);

	g_io_channel_unref(io);
	return TRUE;

out:
	g_string_free(line, TRUE);
	g_io_channel_unref(io);
	return FALSE;
}

VQMailVirtual *v_qmail_virtual_new(GError **error)
{
	struct stat statbuf;
	VQMailVirtual *virtual = NULL;
	GError *my_error = NULL;
	gchar *filename;

	g_return_val_if_fail(!error || !*error, NULL);

	virtual = g_object_new(V_TYPE_QMAIL_VIRTUAL, NULL);

	filename = g_strconcat(v_qmail_get_root(), LOCAL_DOMAINS_FILE, NULL);
	if (g_stat(filename, &statbuf) == 0)
		virtual->ctime_local = statbuf.st_ctime;
	if (!parse_file(virtual->entries, filename, 1, &my_error)) {
		if (my_error->domain == G_FILE_ERROR &&
		    my_error->code == G_FILE_ERROR_NOENT) {
			g_error_free(my_error);
			my_error = NULL;
		} else
			goto out;
	}
	virtual->filename_local = filename;
	virtual->ctime_local = statbuf.st_ctime;

	filename = g_strconcat(v_qmail_get_root(), VIRTUAL_DOMAINS_FILE, NULL);
	if (g_stat(filename, &statbuf) == 0)
		virtual->ctime_virtual = statbuf.st_ctime;
	if (!parse_file(virtual->entries, filename, 2, &my_error)) {
		if (my_error->domain == G_FILE_ERROR &&
		    my_error->code == G_FILE_ERROR_NOENT) {
			g_error_free(my_error);
			my_error = NULL;
		} else
			goto out;
	}
	virtual->filename_virtual = filename;
	virtual->ctime_virtual = statbuf.st_ctime;

	return virtual;
out:
	g_free(filename);
	g_object_unref(virtual);
	g_propagate_error(error, my_error);
	return NULL;
}

VQMailVirtual *v_qmail_virtual_get_default(void)
{
	static time_t last_checked = 0;
	VQMailVirtual *qmail_virtual;
	time_t now = time(NULL);

	G_LOCK(_default_lock);
	if (_default_virtual && now > last_checked) {
		last_checked = now;
		if (v_qmail_virtual_changed(_default_virtual)) {
			g_object_unref(_default_virtual);
			_default_virtual = NULL;
		}
	}
	if (!_default_virtual) {
		last_checked = now;
		_default_virtual = v_qmail_virtual_new(NULL);
	}
	qmail_virtual = g_object_ref(_default_virtual);
	G_UNLOCK(_default_lock);

	return qmail_virtual;
}

void v_qmail_virtual_invalidate_default(void)
{
	G_LOCK(_default_lock);
	if (_default_virtual)
		g_object_unref(_default_virtual);
	_default_virtual = NULL;
	G_UNLOCK(_default_lock);
}

gboolean v_qmail_virtual_changed(VQMailVirtual *qmail_virtual)
{
	struct stat statbuf;

	g_return_val_if_fail(!qmail_virtual ||
	                     V_IS_QMAIL_VIRTUAL(qmail_virtual), FALSE);

	if (!qmail_virtual)
		qmail_virtual = v_qmail_virtual_get_default();

	if (g_stat(qmail_virtual->filename_local, &statbuf) != 0)
		statbuf.st_ctime = 0;
	if (statbuf.st_ctime != qmail_virtual->ctime_local)
		return TRUE;

	if (g_stat(qmail_virtual->filename_virtual, &statbuf) != 0)
		statbuf.st_ctime = 0;
	if (statbuf.st_ctime != qmail_virtual->ctime_virtual)
		return TRUE;

	return FALSE;
}
