#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include <glib.h>
#include <glib/gstdio.h>
#include <glib-object.h>

#include <misc/user.h>
#include <misc/utils.h>
#include <qmail/qmail.h>
#include <vdomain/vdomain.h>
#include "internal.h"

struct _VAccount {
	GObject			gobject;

	VDomainEntry		*entry;
	VMaildir		*maildir;
};

struct _VAccountClass {
	GObjectClass		parent_class;
};

G_DEFINE_TYPE(VAccount, v_account, G_TYPE_OBJECT);

static void v_account_dispose(GObject *gobject);

static void v_account_class_init(VAccountClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->dispose = v_account_dispose;
}

static void v_account_init(VAccount *account)
{
}

static void v_account_dispose(GObject *gobject)
{
	VAccount *account = V_ACCOUNT(gobject);

	if (account->maildir) {
		g_object_unref(account->maildir);
		account->maildir = NULL;
	}

	if (account->entry) {
		g_object_unref(account->entry);
		account->entry = NULL;
	}


	G_OBJECT_CLASS(v_account_parent_class)->dispose(gobject);
}

VAccount *v_account_new(VDomain *vdomain, VDomainEntry *entry, GError **error)
{
	VAccount *account;
	VMaildir *maildir = NULL;

	g_return_val_if_fail(V_IS_DOMAIN(vdomain), NULL);
	g_return_val_if_fail(V_IS_DOMAIN_ENTRY(entry), NULL);
	g_return_val_if_fail(!error || !*error, NULL);

	if (entry->maildir) {
		maildir = v_maildir_new(entry->maildir, error);
		if (!maildir)
			return NULL;
	}

	account = g_object_new(V_TYPE_ACCOUNT, NULL);
	account->entry = g_object_ref(entry);
	account->maildir = maildir;

	return account;
}

VMaildir *v_account_get_maildir(VAccount *account)
{
	g_return_val_if_fail(V_IS_ACCOUNT(account), FALSE);

	return account->maildir ? g_object_ref(account->maildir) : NULL;
}

static inline gboolean v_account_rewind_stream(GIOChannel *io, GError **error)
{
	g_assert(io);
	g_assert(!error || !*error);

	return g_io_channel_seek_position(io, 0, G_SEEK_SET, error)
	       == G_IO_STATUS_NORMAL;
}

gboolean v_account_deliver(VAccount *account, GIOChannel *mail,
                           const gchar *sender, const gchar *domain,
                           gchar **headers, GError **error)
{
	gboolean retval = FALSE;
	gpointer handle = NULL;

	g_return_val_if_fail(V_IS_ACCOUNT(account), FALSE);
	g_return_val_if_fail(mail, FALSE);
	g_return_val_if_fail(sender, FALSE);
	g_return_val_if_fail(domain, FALSE);
	g_return_val_if_fail(!error || !*error, FALSE);

	if (account->maildir && account->entry->enabled) {
		if (!v_account_rewind_stream(mail, error))
			goto out;

		handle = v_maildir_prepare(account->maildir, mail, headers,
		                           error);
		if (!handle)
			goto out;
	}

	if (account->entry->forwards) {
		gboolean result;
		gchar **forwards = account->entry->forwards;
		gchar **recipients;
		int i;

		for(i = 0; forwards[i]; i++);
		recipients = g_new(gchar *, i + 1);
		for(i = 0; forwards[i]; i++) {
			gchar *p = strchr(forwards[i], '@');

			if (p && !*(p + 1)) {
				const gchar *me = v_qmail_me();

				if (!me)
					goto out;
				recipients[i] = g_strconcat(forwards[i],
				                            me, NULL);
			} else if (p)
				recipients[i] = g_strdup(forwards[i]);
			else
				recipients[i] = g_strconcat(forwards[i], "@",
				                            domain, NULL);
		}
		recipients[i] = NULL;
		result = v_qmail_inject(mail, sender, recipients, error);
		g_strfreev(recipients);

		if (!result)
			goto out;
	}

	
	if (!v_account_rewind_stream(mail, error))
		goto out;

	if (handle && !v_maildir_deliver(account->maildir, handle, error))
		goto out;
	handle = NULL;

	retval = TRUE;
out:
	v_account_rewind_stream(mail, NULL);

	if (handle)
		v_maildir_abort(account->maildir, handle);

	return retval;
}

void v_account_set_environent(VAccount *account)
{
	VDomainEntry *entry;
	gchar buf[32];

	g_return_if_fail(V_IS_ACCOUNT(account));

	buf[sizeof buf - 1] = 0;
	entry = account->entry;

	g_setenv("VUSER", entry->name, TRUE);

	if (entry->maildir) {
		gchar *tmp = v_maildir_get_base(account->maildir);

		g_setenv("MAILDIR", tmp, TRUE);
		g_free(tmp);
	}

	g_setenv("VUSER_PERSONAL", entry->personal ? entry->personal : "",
	         TRUE);

	snprintf(buf, sizeof buf - 1, "%lu", (gulong)entry->ctime);
	g_setenv("VUSER_CTIME", buf, TRUE);

	snprintf(buf, sizeof buf - 1, "%lu", (gulong)entry->expiry);
	g_setenv("VUSER_EXPIRY", buf, TRUE);

	snprintf(buf, sizeof buf - 1, "%u", entry->msgcount);
	g_setenv("VUSER_MSGCOUNT", buf, TRUE);

	snprintf(buf, sizeof buf - 1, "%u", entry->msgsize);
	g_setenv("VUSER_MSGSIZE", buf, TRUE);

	snprintf(buf, sizeof buf - 1, "%u", entry->hardquota);
	g_setenv("VUSER_HARDQUOTA", buf, TRUE);

	snprintf(buf, sizeof buf - 1, "%u", entry->softquota);
	g_setenv("VUSER_SOFTQUOTA", buf, TRUE);

	g_setenv("VUSER_MAILBOX_ENABLED", entry->enabled ? "1" : "0", TRUE);
}
