#include <string.h>
#include <stdlib.h>
#include <time.h>

#include <glib.h>
#include <glib/gstdio.h>
#include <glib-object.h>

#include <misc/utils.h>
#include <misc/user.h>
#include <cdb/cdb.h>
#include <vdomain/vdomain.h>
#include "internal.h"

#define DEFAULT_CDB_PATH		"/passwd.cdb"

#define V_TYPE_DOMAIN_CDB		(v_domain_cdb_get_type())
#define V_DOMAIN_CDB(o)			(G_TYPE_CHECK_INSTANCE_CAST((o), V_TYPE_DOMAIN_CDB, VDomainCDB))
#define V_IS_DOMAIN_CDB(o)		(G_TYPE_CHECK_INSTANCE_TYPE((o), V_TYPE_DOMAIN_CDB))
#define V_DOMAIN_CDB_GET_CLASS(inst)	(G_TYPE_INSTANCE_GET_CLASS((inst), V_TYPE_DOMAIN_CDB, VDomainCDBClass))

#define V_DOMAIN_CDB_ITER(i)		((VDomainCDBIter *)(i))

typedef struct _VDomainCDB	VDomainCDB;
typedef struct _VDomainCDBClass	VDomainCDBClass;
typedef struct _VDomainCDBIter	VDomainCDBIter;

enum {
	ATTR_PASS = 1,
	ATTR_DEST,
	ATTR_HARDQUOTA,
	ATTR_SOFTQUOTA,
	ATTR_MSGSIZE,
	ATTR_MSGCOUNT,
	ATTR_EXPIRY,
	ATTR_MAILBOX_ENABLED,
	ATTR_PERSONAL
};

struct _VDomainCDB {
	VDomain		vdomain;

	CDBReader	*cdb;
	GPtrArray	*iters;
};

struct _VDomainCDBClass {
	VDomainClass	parent_class;
};

struct _VDomainCDBIter {
	VDomainIter	iter;

	gchar		*last;
	enum {
		ITER_OK,
		ITER_INVALID,
		ITER_EOF
	}		status;
};

GType v_domain_cdb_get_type(void);

G_DEFINE_TYPE(VDomainCDB, v_domain_cdb, V_TYPE_DOMAIN);

static void v_domain_cdb_dispose(GObject *gobject);
static void v_domain_cdb_finalize(GObject *gobject);

static gboolean v_domain_cdb_setup(VDomain *vdomain, GError **error);
static VDomainEntry *v_domain_cdb_lookup(VDomain *vdomain, const gchar *user,
                                         GError **error);
static gboolean v_domain_cdb_commit(VDomain *vdomain, GError **error);

static void v_domain_cdb_iter_init(VDomainIter *viter);
static VDomainEntry *v_domain_cdb_iter_next(VDomainIter *viter,
                                            GError **error);
static void v_domain_cdb_iter_free(VDomainIter *viter);

static void v_domain_cdb_class_init(VDomainCDBClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
	VDomainClass *vdomain_class = V_DOMAIN_CLASS(klass);

	gobject_class->dispose = v_domain_cdb_dispose;
	gobject_class->finalize = v_domain_cdb_finalize;

	vdomain_class->iter_size = sizeof(VDomainCDBIter);

	vdomain_class->setup = v_domain_cdb_setup;
	vdomain_class->lookup = v_domain_cdb_lookup;
	vdomain_class->commit = v_domain_cdb_commit;

	vdomain_class->iter_init = v_domain_cdb_iter_init;
	vdomain_class->iter_next = v_domain_cdb_iter_next;
	vdomain_class->iter_free = v_domain_cdb_iter_free;
}

static void v_domain_cdb_init(VDomainCDB *domain)
{
}

static void v_domain_cdb_dispose(GObject *gobject)
{
	VDomainCDB *domain = V_DOMAIN_CDB(gobject);

	if (domain->cdb) {
		g_object_unref(domain->cdb);
		domain->cdb = NULL;
	}

	G_OBJECT_CLASS(v_domain_cdb_parent_class)->dispose(gobject);
}

static void v_domain_cdb_finalize(GObject *gobject)
{
	VDomainCDB *domain = V_DOMAIN_CDB(gobject);

	if (domain->iters) {
		g_assert(!domain->iters->len);
		g_ptr_array_free(domain->iters, TRUE);
		domain->iters = NULL;
	}

	G_OBJECT_CLASS(v_domain_cdb_parent_class)->finalize(gobject);
}

static gboolean v_domain_cdb_setup(VDomain *vdomain, GError **error)
{
	VDomainCDB *domain = V_DOMAIN_CDB(vdomain);
	gchar *path;

	g_return_val_if_fail(vdomain->user, FALSE);

	path = g_strconcat(vdomain->user->home, DEFAULT_CDB_PATH, NULL);
	if (g_file_test(path, G_FILE_TEST_EXISTS))
		domain->cdb = cdb_reader_new(path, error);
	g_free(path);

	if (!domain->cdb)
		return FALSE;

	return TRUE;
}

#define DECODE_TEMPLATE(type, max, decoder, dectype)			\
	static inline type decode_##type(gchar **ptr)			\
	{								\
		type val;						\
									\
		if (**ptr == '-')					\
			val = max;					\
		else {							\
			dectype tmp = decoder(*ptr, NULL, 10);		\
									\
			val = MIN(max, tmp); 		 		\
		}							\
		if (val < 0)						\
			val = max;					\
									\
		*ptr += strlen(*ptr) + 1;				\
									\
		return val;						\
	}

DECODE_TEMPLATE(guint, G_MAXUINT, strtoul, gulong)
DECODE_TEMPLATE(time_t, TIME_T_MAX, strtol, glong)

#undef DECODE_TEMPLATE

static VDomainEntry *v_domain_cdb_entry_decode(VDomainCDB *domain,
                                               CDBDatum *datum, GError **error)
{
	VDomainEntry *entry = NULL;
	gboolean enabled = TRUE;
	GError *my_error = NULL;
	gchar *pass, *mailbox, *personal, **forwards = NULL;
	gchar *p, *start, *end, *tmp = NULL;
	guint hardquota, softquota, msgsize, msgcount;
	time_t ctime, expiry;
	int i;

	p = datum->data;
	end = datum->data + datum->dlen;

	if ((i = *p++) != 2) {
		g_set_error(&my_error, V_DOMAIN_ERROR, V_DOMAIN_ERROR_UNSUPP,
		            _("Unsupported VPW record version %d\n"), i);
		goto out;
	}

	while(p < end) {
		guchar value, flag = *p++;

		if (!flag)
			break;
		if (p >= end)
			goto out;
		value = *p++;

		switch(flag) {
		case ATTR_MAILBOX_ENABLED:
			enabled = !!value;
			break;
		default:
			goto out;
		}
	}

	pass = p;
	p += strlen(pass);
	if (p++ >= end)
		goto out;

	mailbox = p;
	p += strlen(mailbox);
	if (p++ >= end)
		goto out;

	for(start = p, i = 0; p < end && *p; i++)
		p += strlen(p) + 1;
	if (i) {
		forwards = g_new(gchar *, i + 1);
		for(i = 0; start < p; i++) {
			forwards[i] = start;
			start += strlen(start) + 1;
		}
		forwards[i] = NULL;
	}
	if (p++ >= end)
		goto out;

	personal = p;
	p += strlen(personal) + 1;
	if (p >= end)
		goto out;

	hardquota = decode_guint(&p);
	if (p >= end)
		goto out;

	softquota = decode_guint(&p);
	if (p >= end)
		goto out;

	msgsize = decode_guint(&p);
	if (p >= end)
		goto out;

	msgcount = decode_guint(&p);
	if (p >= end)
		goto out;

	ctime = decode_time_t(&p);
	if (p >= end)
		goto out;

	expiry = decode_time_t(&p);
	if (p > end)
		goto out;

	if (mailbox[0]) {
		while(mailbox[0] == '.' && mailbox[1] == '/')
			while(*++mailbox == '/');
		if (mailbox[0] != '/')
			tmp = g_strconcat(V_DOMAIN(domain)->user->home, "/",
			                  mailbox, NULL);
	} else
		mailbox = NULL;

	entry = v_domain_entry_new(datum->key, tmp ? tmp : mailbox);
	v_domain_entry_set_password(entry, pass);
	v_domain_entry_set_forwards_v(entry, forwards);
	v_domain_entry_set_personal(entry, personal);
	entry->hardquota = hardquota;
	entry->softquota = softquota;
	entry->msgsize = msgsize;
	entry->msgcount = msgcount;
	entry->ctime = ctime;
	entry->expiry = expiry;
	entry->enabled = enabled;

out:
	g_free(tmp);
	g_free(forwards);

	if (my_error)
		g_propagate_error(error, my_error);
	else if (!entry)
		g_set_error(error, V_DOMAIN_ERROR, V_DOMAIN_ERROR_INVALID,
		            _("Invalid VPW record"));

	return entry;
}

#define ENCODE_TEMPLATE(type, max, format, formattype) \
	static inline void encode_##type(GString *data, type value)	\
	{								\
		if (value >= max)					\
			g_string_append_c(data, '-');			\
		else							\
			g_string_append_printf(data,			\
					"%" format, (formattype)value);	\
		g_string_append_c(data, 0);				\
	}

ENCODE_TEMPLATE(guint, G_MAXUINT, "u", guint)
ENCODE_TEMPLATE(time_t, TIME_T_MAX, TIME_T_FORMAT, TIME_T_GLIB_TYPE)

#undef ENCODE_TEMPLATE

static CDBDatum *v_domain_cdb_entry_encode(VDomainCDB *domain,
                                           VDomainEntry *entry)
{
	CDBDatum *datum;
	GString *data = g_string_new(NULL);

	g_string_append_c(data, 2);
	g_string_append_c(data, ATTR_MAILBOX_ENABLED);
	g_string_append_c(data, entry->enabled);
	g_string_append_c(data, 0);

	if (entry->passwd)
		g_string_append(data, entry->passwd);
	g_string_append_c(data, 0);

	if (entry->maildir) {
		const gchar *home = V_DOMAIN(domain)->user->home;
		int len = strlen(home);

		if (strncmp(entry->maildir, home, len) == 0 &&
		    entry->maildir[len] == '/') {
		    	g_string_append_c(data, '.');
			g_string_append(data, entry->maildir + len);
		} else
			g_string_append(data, entry->maildir);
	}
	g_string_append_c(data, 0);

	if (entry->forwards) {
		gchar **forward;

		for(forward = entry->forwards; *forward; forward++) {
			g_string_append(data, *forward);
			g_string_append_c(data, 0);
		}
	}
	g_string_append_c(data, 0);

	if (entry->personal)
		g_string_append(data, entry->personal);
	g_string_append_c(data, 0);

	encode_guint(data, entry->hardquota);
	encode_guint(data, entry->softquota);
	encode_guint(data, entry->msgsize);
	encode_guint(data, entry->msgcount);
	encode_time_t(data, entry->ctime);
	encode_time_t(data, entry->expiry);

	datum = cdb_datum_new(strlen(entry->name), g_strdup(entry->name),
	                      data->len, data->str, FALSE);

	g_string_free(data, FALSE);

	return datum;
}

static void v_domain_cdb_invalidate_iters(VDomainCDB *domain)
{
	int i;

	if (!domain->iters)
		return;

	for(i = 0; i < domain->iters->len; i++) {
		VDomainCDBIter *iter = g_ptr_array_index(domain->iters, i);

		if (iter->status != ITER_EOF)
			iter->status = ITER_INVALID;
	}
}

static VDomainEntry *v_domain_cdb_lookup(VDomain *vdomain, const gchar *user,
                                         GError **error)
{
	VDomainEntry *entry = NULL;
	VDomainCDB *domain = V_DOMAIN_CDB(vdomain);
	CDBDatum *datum;
	GError *my_error = NULL;

	if (!domain->cdb)
		goto out;

	v_domain_cdb_invalidate_iters(domain);

	datum = cdb_reader_get_string(domain->cdb, user, &my_error);
	if (datum) {
		entry = v_domain_cdb_entry_decode(domain, datum, &my_error);
		cdb_datum_free(datum);
	}

out:
	if (my_error)
		g_propagate_error(error, my_error);
	else if (!entry)
		g_set_error(error, V_DOMAIN_ERROR, V_DOMAIN_ERROR_NOENT,
		            _("Unknown user `%s'."), user);

	return entry;
}

struct commit_user_data {
	VDomainCDB	*domain;
	CDBWriter	*writer;
	GError		*error;
};

static void write_entry(gpointer key, gpointer value, gpointer user_data)
{
	struct commit_user_data *data = user_data;
	CDBDatum *datum;

	if (data->error || !value)
		return;

	datum = v_domain_cdb_entry_encode(data->domain,
	                                  V_DOMAIN_ENTRY(value));
	cdb_writer_putrec(data->writer, datum, &data->error);
	cdb_datum_free(datum);
}

static gboolean v_domain_cdb_commit(VDomain *vdomain, GError **error)
{
	struct commit_user_data data;
 	VDomainCDB *domain = V_DOMAIN_CDB(vdomain);
	CDBWriter *writer = NULL;
	CDBDatum *datum = NULL;
	gboolean retval = FALSE;
	GError *my_error = NULL;
	gchar *path = NULL, *tmp;

	v_domain_cdb_invalidate_iters(domain);

	if (domain->cdb && !cdb_reader_firstrec(domain->cdb, &my_error))
		goto out;

	path = g_strconcat(vdomain->user->home, DEFAULT_CDB_PATH, NULL);

	tmp = strrchr(path, '.');
	if (tmp && !strchr(path, '/')) {
		gchar *p = tmp;

		*p = 0;
		tmp = g_strconcat(path, ".tmp", NULL);
		*p = '.';
	} else
		tmp = g_strconcat(path, ".tmp", NULL);
	writer = cdb_writer_new(tmp, 0600, &my_error);
	g_free(tmp);
	if (!writer)
		goto out;

	while(domain->cdb &&
	      (datum = cdb_reader_nextrec(domain->cdb, &my_error)) != NULL) {
		VDomainEntry *entry;

		entry = v_domain_cdb_entry_decode(domain, datum, &my_error);
		if (!entry ||
		    !g_hash_table_lookup_extended(vdomain->changeset,
		                                  entry->name, NULL, NULL)) {
			if (!cdb_writer_putrec(writer, datum, &my_error))
				goto out;
		}

		if (entry) {
			g_object_unref(entry);
			entry = NULL;
		}
		cdb_datum_free(datum);
	}

	data.domain = domain;
	data.writer = writer;
	data.error = NULL;
	g_hash_table_foreach(vdomain->changeset, write_entry, &data);
	if (data.error) {
		my_error = data.error;
		goto out;
	}

	if (!cdb_writer_finish(writer, path, &my_error))
		goto out;

	retval = TRUE;
out:
	if (datum)
		cdb_datum_free(datum);
	if (writer)
		g_object_unref(writer);
	g_free(path);

	if (my_error)
		g_propagate_error(error, my_error);

	return retval;
}

static void v_domain_cdb_iter_init(VDomainIter *viter)
{
	VDomainCDBIter *iter = V_DOMAIN_CDB_ITER(viter);
	VDomainCDB *domain = V_DOMAIN_CDB(viter->vdomain);

	iter->status = domain->cdb ? ITER_INVALID : ITER_EOF;

	if (!domain->iters)
		domain->iters = g_ptr_array_new();

	g_ptr_array_add(domain->iters, iter);
}

static VDomainEntry *v_domain_cdb_iter_next(VDomainIter *viter, GError **error)
{
	VDomainCDBIter *iter = V_DOMAIN_CDB_ITER(viter);
	VDomainEntry *entry = NULL;
	VDomainCDB *domain = V_DOMAIN_CDB(viter->vdomain);
	CDBDatum *datum;
	GError *my_error = NULL;

	if (iter->status == ITER_EOF)
		return NULL;

	if (iter->status == ITER_INVALID) {
		if (iter->last) {
			datum = cdb_reader_get_string(domain->cdb, iter->last,
			                              &my_error);
			if (!datum)
				goto out;
			cdb_datum_free(datum);
		} else {
			if (!cdb_reader_firstrec(domain->cdb, &my_error))
				goto out;
		}
	}

retry:
	datum = cdb_reader_nextrec(domain->cdb, &my_error);

	g_free(iter->last);
	iter->last = NULL;

	if (datum) {
		iter->status = ITER_OK;
		iter->last = g_strdup(datum->key);
		entry = v_domain_cdb_entry_decode(domain, datum, &my_error);
		cdb_datum_free(datum);

		/* this record could somehow not be decoded */
		if (!entry) {
			g_error_free(my_error);
			my_error = NULL;
			goto retry;
		}
	} else
		iter->status = ITER_EOF;

out:
	if (my_error)
		g_propagate_error(error, my_error);

	return entry;
}

static void v_domain_cdb_iter_free(VDomainIter *viter)
{
	VDomainCDBIter *iter = V_DOMAIN_CDB_ITER(viter);
	VDomainCDB *domain = V_DOMAIN_CDB(viter->vdomain);

	if (iter->last) {
		g_free(iter->last);
		iter->last = NULL;
	}

	g_assert(domain->iters);
	g_assert(g_ptr_array_remove_fast(domain->iters, iter));
}
