#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>

#include <glib.h>
#include <glib/gstdio.h>
#include <glib-object.h>

#include <misc/utils.h>
#include <misc/crc32.h>
#include <misc/user.h>
#include <misc/vconfig.h>
#include <qmail/qmail.h>
#include <vdomain/vdomain.h>
#include "internal.h"

#define DEFAULT_MAILBOX_PATH	"%H/users/%S/%U"
#define DEFAULT_USERDIR_SLICES	0
#define DEFAULT_USERDIR_BITS	0

#define NO_ESCAPE_LIST		"abcdefghijklmnopqrstuvwxyz0123456789-_^~"

GQuark v_domain_error_quark(void)
{
	static GQuark error_quark = 0;

	if (G_UNLIKELY(error_quark == 0))
		error_quark =
			g_quark_from_static_string("v-domain-error-quark");

	return error_quark;
}

G_DEFINE_ABSTRACT_TYPE(VDomain, v_domain, G_TYPE_OBJECT);

static void v_domain_dispose(GObject *gobject);
static void v_domain_finalize(GObject *gobject);

static void v_domain_class_init(VDomainClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->dispose = v_domain_dispose;
	gobject_class->finalize = v_domain_finalize;
}

static void v_domain_init(VDomain *vdomain)
{
}

static void v_domain_dispose(GObject *gobject)
{
	VDomain *vdomain = V_DOMAIN(gobject);

	if (vdomain->user) {
		g_object_unref(vdomain->user);
		vdomain->user = NULL;
	}

	if (vdomain->changeset) {
		g_hash_table_destroy(vdomain->changeset);
		vdomain->changeset = NULL;
	}

	G_OBJECT_CLASS(v_domain_parent_class)->dispose(gobject);
}

static void v_domain_finalize(GObject *gobject)
{
	VDomain *vdomain = V_DOMAIN(gobject);

	g_free(vdomain->domain);
	vdomain->domain = NULL;

	G_OBJECT_CLASS(v_domain_parent_class)->finalize(gobject);
}

static void v_domain_initialize_types(void)
{
	static gboolean initialized = FALSE;

	if (initialized)
		return;
	initialized = TRUE;

	v_absorb(v_domain_cdb_get_type());
	v_absorb(v_domain_cdb_get_type());
}

VDomain *v_domain_new(const gchar *type, const gchar *domain, gboolean create,
                      GError **error)
{
	VQMailVirtual *virtual = NULL;
	const gchar *vuser = NULL;
	VDomain *vdomain = NULL;
	GError *my_error = NULL;
	GType gtype;
	gchar *tmp;

	g_return_val_if_fail(!error || !*error, NULL);

	v_domain_initialize_types();

	if (!type)
		type = v_config_get_string(NULL, NULL, "domain-handler");
	if (!type) {
		g_set_error(&my_error, V_DOMAIN_ERROR, V_DOMAIN_ERROR_INVALID,
		            _("No domain handler specified."));
		goto out;
	}

	tmp = g_strconcat("VDomain", type, NULL);
	gtype = g_type_from_name(tmp);
	g_free(tmp);

	if (!g_type_is_a(gtype, V_TYPE_DOMAIN) || G_TYPE_IS_ABSTRACT(gtype)) {
		g_set_error(&my_error, V_DOMAIN_ERROR, V_DOMAIN_ERROR_INVALID,
		            _("Unknown virtual domain type `%s'."), type);
		goto out;
	}

	if (domain) {
		virtual = v_qmail_virtual_get_default();
		if (!virtual) {
			g_set_error(&my_error, V_DOMAIN_ERROR,
			            V_DOMAIN_ERROR_INVALID,
			            _("Virtual domain configuration "
			              "missing or invalid."));
			goto out;
		}
		tmp = g_strstrip(g_ascii_strdown(domain, -1));
		vuser = v_qmail_virtual_lookup(virtual, tmp);
		if (!vuser) {
			g_set_error(&my_error, V_DOMAIN_ERROR,
			            V_DOMAIN_ERROR_NOENT,
			            _("Unknown virtual domain `%s'."), tmp);
			g_free(tmp);
			goto out;
		}
	} else
		tmp = NULL;

	vdomain = g_object_new(gtype, NULL);
	g_assert(vdomain);

	vdomain->domain = tmp;
	vdomain->user = v_user_new(vuser);

	if (!vdomain->user) {
		g_set_error(&my_error, V_DOMAIN_ERROR, V_DOMAIN_ERROR_INVALID,
		            _("Underlying user `%s' invalid."), vuser);
		goto out;
	}

	if (!V_DOMAIN_GET_CLASS(vdomain)->setup(vdomain, &my_error)) {
		if (create && my_error->domain == V_DOMAIN_ERROR &&
		    my_error->code == V_DOMAIN_ERROR_NOENT) {
			g_error_free(my_error);
			my_error = NULL;
		} else
			goto out;
	}

out:
	if (my_error) {
		g_propagate_error(error, my_error);
		if (vdomain) {
			g_object_unref(vdomain);
			vdomain = NULL;
		}
	}
	if (virtual)
		g_object_unref(virtual);

	return vdomain;
}

void v_domain_set_auth_method(VDomain *vdomain, GType auth_type)
{
	g_return_if_fail(V_IS_DOMAIN(vdomain));
	g_return_if_fail(g_type_is_a(auth_type, V_TYPE_DOMAIN_AUTH));
	g_return_if_fail(g_type_is_a(auth_type, G_TYPE_OBJECT));
	g_return_if_fail(!G_TYPE_IS_ABSTRACT(auth_type));

	vdomain->auth_type = auth_type;
}

VDomainEntry *v_domain_get_entry(VDomain *vdomain, const gchar *account,
                                 GError **error)
{
	VDomainEntry *entry;
	gchar *tmp;

	g_return_val_if_fail(V_IS_DOMAIN(vdomain), NULL);
	g_return_val_if_fail(account, NULL);
	g_return_val_if_fail(!error || !*error, NULL);

	tmp = g_strstrip(g_ascii_strdown(account, -1));
	entry = V_DOMAIN_GET_CLASS(vdomain)->lookup(vdomain, tmp, error);
	g_free(tmp);

	return entry;
}

VAccount *v_domain_get_account(VDomain *vdomain, const gchar *account,
                               GError **error)
{
	VDomainEntry *entry;

	g_return_val_if_fail(V_IS_DOMAIN(vdomain), NULL);
	g_return_val_if_fail(account, NULL);
	g_return_val_if_fail(!error || !*error, NULL);

	entry = v_domain_get_entry(vdomain, account, error);
	if (!entry)
		return NULL;

	return v_account_new(vdomain, entry, error);
}

static VDomainEntry *
v_domain_auth_internal(VDomain *vdomain, const gchar *account,
                       const gchar *pass, GError **error)
{
	VDomainAuthClass *klass;
	VDomainEntry *entry = NULL;
	VDomainAuth *auth = NULL;
	gboolean result = FALSE;
	GError *my_error = NULL;

	g_return_val_if_fail(!error || !*error, NULL);

	entry = v_domain_get_entry(vdomain, account, &my_error);
	if (!entry) {
		if (my_error->domain == V_DOMAIN_ERROR &&
		    my_error->code == V_DOMAIN_ERROR_NOENT) {
			g_error_free(my_error);
			my_error = NULL;
			g_set_error(error, V_AUTH_ERROR, V_AUTH_ERROR_NOENT,
			            _("Invalid or unknown virtual user"));
		} else
			g_propagate_error(error, my_error);
		goto out;
	}

	if (time(NULL) >= entry->expiry) {
		g_set_error(error, V_AUTH_ERROR, V_AUTH_ERROR_NOENT,
		            _("Account has expired"));
		goto out;
	}

	if (!vdomain->auth_type ||
	    !(auth = g_object_new(vdomain->auth_type, NULL))) {
		g_set_error(error, V_AUTH_ERROR, V_AUTH_ERROR_NOTSUPP,
		            _("No authentication method set"));
		goto out;
	}

	klass = V_DOMAIN_AUTH_GET_CLASS(auth);
	klass->setup(auth, entry);
	result = klass->auth(auth, pass, error);

out:
	if (!result && entry) {
		g_object_unref(entry);
		entry = NULL;
	}
	if (auth)
		g_object_unref(auth);

	return entry;
}

gboolean v_domain_auth(VDomain *vdomain, const gchar *account,
                       const gchar *pass, GError **error)
{
	VDomainEntry *entry;

	g_return_val_if_fail(V_IS_DOMAIN(vdomain), FALSE);
	g_return_val_if_fail(account, FALSE);
	g_return_val_if_fail(!error || !*error, FALSE);

	entry = v_domain_auth_internal(vdomain, account, pass, error);
	if (!entry)
		return FALSE;

	g_object_unref(entry);
	return TRUE;
}

VAccount *v_domain_auth_account(VDomain *vdomain, const gchar *account,
                                const gchar *pass, GError **error)
{
	VDomainEntry *entry;
	VAccount *vaccount = NULL;
	GError *my_error = NULL;

	g_return_val_if_fail(V_IS_DOMAIN(vdomain), NULL);
	g_return_val_if_fail(account, NULL);
	g_return_val_if_fail(!error || !*error, NULL);

	entry = v_domain_auth_internal(vdomain, account, pass, error);
	if (!entry)
		return FALSE;

	vaccount = v_account_new(vdomain, entry, &my_error);
	if (!vaccount) {
		if (my_error->domain == V_DOMAIN_ERROR &&
		    my_error->code == V_DOMAIN_ERROR_NOENT) {
			g_error_free(my_error);
			my_error = NULL;
			g_set_error(error, V_AUTH_ERROR, V_AUTH_ERROR_NOENT,
			            _("Invalid or unknown virtual user"));
		} else
			g_propagate_error(error, my_error);
	}

	g_object_unref(entry);

	return vaccount;
}

VUser *v_domain_get_user(VDomain *vdomain)
{
	g_return_val_if_fail(V_IS_DOMAIN(vdomain), NULL);

	return g_object_ref(vdomain->user);
}

VDomainIter *v_domain_iter_new(VDomain *vdomain)
{
	VDomainClass *klass;
	VDomainIter *iter;

	g_return_val_if_fail(V_IS_DOMAIN(vdomain), NULL);
	klass = V_DOMAIN_GET_CLASS(vdomain);

	iter = g_malloc0(klass->iter_size);
	iter->vdomain = g_object_ref(vdomain);

	klass->iter_init(iter);

	return iter;
}

VDomainEntry *v_domain_iter_next(VDomainIter *iter, GError **error)
{
	g_return_val_if_fail(iter, NULL);
	g_return_val_if_fail(iter->vdomain, NULL);

	g_return_val_if_fail(!error || !*error, NULL);

	return V_DOMAIN_GET_CLASS(iter->vdomain)->iter_next(iter, error);
}

void v_domain_iter_free(VDomainIter *iter)
{
	g_return_if_fail(iter);
	g_return_if_fail(iter->vdomain);

	V_DOMAIN_GET_CLASS(iter->vdomain)->iter_free(iter);

	g_object_unref(iter->vdomain);
	iter->vdomain = NULL;

	g_free(iter);
}

static void v_domain_entry_unref(gpointer data)
{
	if (data)
		g_object_unref(data);
}

static void v_domain_change(VDomain *vdomain, const gchar *name,
                            VDomainEntry *entry)
{
	if (!vdomain->changeset)
		vdomain->changeset =
			g_hash_table_new_full(g_str_hash, g_str_equal,
			                      g_free, v_domain_entry_unref);

	g_hash_table_insert(vdomain->changeset, g_strdup(name),
	                    entry ? g_object_ref(entry) : NULL);
}

static VDomainEntry *v_domain_lookup(VDomain *vdomain, const gchar *name,
                                     GError **error)
{
	gpointer data;

	g_return_val_if_fail(!error || !*error, NULL);

	if (vdomain->changeset &&
	    g_hash_table_lookup_extended(vdomain->changeset, name,
	                                 NULL, &data)) {
		if (data)
			return g_object_ref(V_DOMAIN_ENTRY(data));
		g_set_error(error, V_DOMAIN_ERROR, V_DOMAIN_ERROR_NOENT,
		            _("Unknown virtual user `%s'."), name);
	}

	return V_DOMAIN_GET_CLASS(vdomain)->lookup(vdomain, name, error);
}

gboolean v_domain_add_entry(VDomain *vdomain, VDomainEntry *entry,
                            GError **error)
{
	VDomainEntry *old;
	GError *my_error = NULL;

	g_return_val_if_fail(V_IS_DOMAIN(vdomain), FALSE);
	g_return_val_if_fail(V_IS_DOMAIN_ENTRY(entry), FALSE);
	g_return_val_if_fail(!error || !*error, FALSE);

	old = v_domain_lookup(vdomain, entry->name, &my_error);
	if (old) {
		g_set_error(error, V_DOMAIN_ERROR, V_DOMAIN_ERROR_EXIST,
		            _("User already exists"));
		g_object_unref(old);
		return FALSE;
	} else
		g_error_free(my_error);

	v_domain_change(vdomain, entry->name, entry);
	return TRUE;
}

gboolean v_domain_modify_entry(VDomain *vdomain, VDomainEntry *entry,
                               GError **error)
{
	VDomainEntry *old;

	g_return_val_if_fail(V_IS_DOMAIN(vdomain), FALSE);
	g_return_val_if_fail(V_IS_DOMAIN_ENTRY(entry), FALSE);
	g_return_val_if_fail(!error || !*error, FALSE);

	old = v_domain_lookup(vdomain, entry->name, error);
	if (!old)
		return FALSE;
	g_object_unref(old);

	v_domain_change(vdomain, entry->name, entry);
	return TRUE;
}

gboolean v_domain_remove_entry(VDomain *vdomain, const gchar *account,
                               GError **error)
{
	VDomainEntry *old;
	gboolean retval = FALSE;
	gchar *tmp;

	g_return_val_if_fail(V_IS_DOMAIN(vdomain), FALSE);
	g_return_val_if_fail(!error || !*error, FALSE);

	tmp = g_strstrip(g_ascii_strdown(account, -1));
	old = v_domain_lookup(vdomain, tmp,error);
	if (!old)
		goto out;
	g_object_unref(old);

	v_domain_change(vdomain, tmp, NULL);
	retval = TRUE;
out:
	g_free(tmp);
	return retval;
}

void v_domain_abort_changes(VDomain *vdomain)
{
	g_return_if_fail(V_IS_DOMAIN(vdomain));

	if (vdomain->changeset) {
		g_hash_table_destroy(vdomain->changeset);
		vdomain->changeset = NULL;
	}
}

gboolean v_domain_commit_changes(VDomain *vdomain, GError **error)
{
	gboolean retval;

	g_return_val_if_fail(V_IS_DOMAIN(vdomain), FALSE);
	g_return_val_if_fail(!error || !*error, FALSE);

	if (!vdomain->changeset)
		return TRUE;

	retval = V_DOMAIN_GET_CLASS(vdomain)->commit(vdomain, error);
	if (retval) {
		g_hash_table_destroy(vdomain->changeset);
		vdomain->changeset = NULL;
	}

	return retval;
}

static gchar *canonicalize_path(const gchar *arg)
{
	GString *retval = g_string_sized_new(strlen(arg) + 4);

	for(; *arg; arg++) {
		guchar p = *arg;

		if (p >= 'A' && p <= 'Z')
			p += 'a' - 'A';

		if (strchr(NO_ESCAPE_LIST, p)) {
			g_string_append_c(retval, p);
			continue;
		}

		switch(p) {
		case ' ':
			g_string_append_c(retval, '+');
			break;
		case '.':
			g_string_append_c(retval, ':');
			break;
		default:
			g_string_append_printf(retval, "=%02x", p);
			break;
		}
	}

	return g_string_free(retval, FALSE);
}

static gchar *slice_path(const gchar *arg)
{
	gchar *retval, *p;
	guint32 hash = v_crc32_string(arg);
	int slices = v_config_get_int(NULL, "defaults", "userdir-slices");
	int bits = v_config_get_int(NULL, "defaults", "userdir-bits");
	int digits;

	if (slices < 0)
		slices = DEFAULT_USERDIR_SLICES;
	if (bits < 0)
		bits = DEFAULT_USERDIR_BITS;

	digits = (bits + 3) / 4;
	retval = g_new(gchar, MAX(1, slices * (digits + 1)));

	for(p = retval; slices > 0; slices--, hash >>= bits) {
		unsigned slice = hash & ((1 << bits) - 1);
		int i;

		for(i = digits; i > 0; i--, slice >>= 4)
			p[i - 1] = "0123456789abcdef"[slice & 15];

		p += digits;
		if (slices > 1)
			*p++ = '/';
	}
	*p = 0;

	return retval;
}

void v_domain_set_entry_password(VDomain *vdomain, VDomainEntry *entry,
                                 const gchar *pass)
{
	VDomainAuthClass *klass;
	VDomainAuth *auth = NULL;

	g_return_if_fail(V_IS_DOMAIN(vdomain));
	g_return_if_fail(V_IS_DOMAIN_ENTRY(entry));

	if (!pass || !vdomain->auth_type ||
	    !(auth = g_object_new(vdomain->auth_type, NULL))) {
		v_domain_entry_set_password(entry, NULL);
		return;
	}

	klass = V_DOMAIN_AUTH_GET_CLASS(auth);
	klass->setup(auth, entry);
	klass->set(auth, pass);

	g_object_unref(auth);
}

gchar *v_domain_build_userdir(VDomain *domain, const gchar *name)
{
	GString *retval;
	gchar *path, *tmp, *p;
	gsize i;

	g_return_val_if_fail(V_IS_DOMAIN(domain), NULL);

	retval = g_string_new(NULL);

	path = v_config_get_string(NULL, "defaults", "maildir");
	if (!path)
		path = DEFAULT_MAILBOX_PATH;
	for(p = path; *p; p++) {
		if (*p == '%') {
			switch(*(++p)) {
			case '%':
				g_string_append_c(retval, '%');
				break;
			case 'D':
				tmp = canonicalize_path(domain->user->user);
				g_string_append(retval, tmp);
				g_free(tmp);
				break;
			case 'H':
				g_string_append(retval, domain->user->home);
				break;
			case 'S':
				if (!name)
					goto out;
				tmp = slice_path(name);
				g_string_append(retval, tmp);
				g_free(tmp);
				break;
			case 'U':
				if (!name)
					goto out;
				tmp = canonicalize_path(name);
				g_string_append(retval, tmp);
				g_free(tmp);
				break;
			}
		} else
			g_string_append_c(retval, *p);
	}

out:
	for(i = 0; i < retval->len; i++) {
		while(retval->str[i] == '/' &&
		      (retval->str[i + 1] == '/' ||
		       retval->str[i + 1] == 0))
			g_string_erase(retval, i, 1);
	}

	return g_string_free(retval, FALSE);
}
