#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include <glib.h>
#include <glib/gstdio.h>
#include <glib-object.h>

#include <misc/utils.h>
#include <misc/vconfig.h>
#include <vdomain/vdomain.h>

#define DEFAULT_POSTMASTER_EMAIL	"postmaster@"
#define DEFAULT_POSTMASTER_ALIASES	default_aliases
#define DEFAULT_QMAIL_HANDLER_FILE	".qmail-default"
#define DEFAULT_QMAIL_HANDLER		"|" BINDIR "/vdeliver"

static const gchar * const default_aliases[] = {
	"postmaster",
	"mailer-daemon",
	"root",
	NULL
};

static gboolean quiet = FALSE;

static const GOptionEntry entries[] = {
	{ "quiet",		0,   0, G_OPTION_ARG_NONE,		&quiet,		"Suppress all status messages",			NULL	},
	{ NULL }
};

static gboolean setup_usersdir(VDomain *domain)
{
	gboolean retval = FALSE;
	gchar *tmp = NULL;

	tmp = v_domain_build_userdir(domain, NULL);
	if (g_file_test(tmp, G_FILE_TEST_EXISTS)) {
		if (!quiet)
			printf(_("%s: Users directory already exists.\n"),
			       g_get_prgname());
	} else {
		GError *error = NULL;

		if (!v_mkdir_recursive(tmp, 0755, 0755, &error)) {
			if (!quiet)
				printf(_("%s: Could not create users "
				         "directory: %s\n"), g_get_prgname(),
				       error->message);
			g_error_free(error);
			goto out;
		}

		if (!quiet)
			printf(_("%s: Successfully created users "
			         "directory.\n"), g_get_prgname());
	}

	retval = TRUE;
out:
	g_free(tmp);
	return retval;
}

static gboolean setup_aliases(VDomain *domain)
{
	VDomainEntry *entry = NULL;
	gboolean retval = FALSE;
	GError *error = NULL;
	gchar **aliases, **alias;
	gchar *email;
	gchar *tmp;

	email = v_config_get_string(NULL, NULL, "postmaster-email");
	if (!email)
		email = DEFAULT_POSTMASTER_EMAIL;
	aliases = v_config_get_strv(NULL, NULL, "postmaster-aliases");
	if (!aliases)
		aliases = (gchar **)DEFAULT_POSTMASTER_ALIASES;

	for(alias = aliases; *alias; alias++) {
		tmp = g_strstrip(g_ascii_strdown(*alias, -1));
		if (!entry) {
			entry = v_domain_entry_new(tmp, NULL);
			v_domain_entry_load_defaults(entry);
			v_domain_entry_set_password(entry, "*");
			v_domain_entry_set_forwards(entry, email, NULL);
		} else {
			VDomainEntry *old = entry;

			entry = v_domain_entry_copy(entry);
			g_object_unref(old);
			v_domain_entry_set_name(entry, tmp);
		}
		g_free(tmp);

		if (!v_domain_add_entry(domain, entry, &error)) {
			if (error->domain != V_DOMAIN_ERROR ||
			    error->code != V_DOMAIN_ERROR_EXIST) {
				if (!quiet)
				    	fprintf(stderr,
				    	        _("%s: Adding alias `%s' "
			    	                  "failed: %s\n"),
				    	        g_get_prgname(), entry->name,
				    	        error->message);
				g_error_free(error);
				goto out;
			}
			g_error_free(error);
			error = NULL;

			if (!quiet)
				g_warning(_("User `%s' already exists, "
				            "skipping."), entry->name);
		} else
			fprintf(stderr, _("%s: Added postmaster alias "
			                  "`%s'.\n"),
			        g_get_prgname(), entry->name);
	}

	retval = TRUE;
out:
	if (entry)
		g_object_unref(entry);
	if (email != DEFAULT_POSTMASTER_EMAIL)
		g_free(email);
	if (aliases != (gchar **)DEFAULT_POSTMASTER_ALIASES)
		g_strfreev(aliases);

	return TRUE;
}

static gboolean setup_qmail_handler(VDomain *domain)
{
	GIOChannel *io = NULL;
	GIOStatus status;
	gboolean renamed = FALSE;
	gboolean retval = FALSE;
	GError *error = NULL;
	VUser *user;
	gchar *path, *tmp;

	user = v_domain_get_user(domain);
	path = g_strconcat(user->home,
	                   "/" DEFAULT_QMAIL_HANDLER_FILE, NULL);
	g_object_unref(user);

	if (g_file_test(path, G_FILE_TEST_EXISTS)) {
		gboolean result;

		if (!quiet)
			g_warning(_("`%s' file exists, renaming to `%s'."),
			          DEFAULT_QMAIL_HANDLER_FILE,
			          DEFAULT_QMAIL_HANDLER_FILE "~");

		tmp = g_strconcat(path, "~", NULL);		
		result = (g_rename(path, tmp) == 0);
		g_free(tmp);

		if (!result) {
			if (!quiet)
				fprintf(stderr,
				        _("%s: Error: Rename failed: %s\n"),
				        g_get_prgname(),
				        g_strerror(errno));
			goto out;
		}
		renamed = TRUE;
	}

	io = v_output_io_channel_new(path, V_IO_FLAGS_CREATE | V_IO_FLAGS_EXCL,
	                             0644, &error);
	if (!io)
		goto out;
	g_io_channel_set_encoding(io, NULL, NULL);

	status = g_io_channel_write_chars(io, DEFAULT_QMAIL_HANDLER "\n",
	                                  sizeof DEFAULT_QMAIL_HANDLER,
	                                  NULL, &error);
	if (status != G_IO_STATUS_NORMAL)
		goto out;

	status = g_io_channel_flush(io, &error);
	if (status != G_IO_STATUS_NORMAL)
		goto out;

	if (!quiet)
		fprintf(stderr, _("%s: Wrote `%s' file.\n"),
		       g_get_prgname(), DEFAULT_QMAIL_HANDLER_FILE);

	retval = TRUE;
out:
	if (error) {
		fprintf(stderr, _("%s: Error creating `%s': %s\n"),
		        g_get_prgname(), path, error->message);
		g_error_free(error);
	}
	if (io) {
		g_io_channel_unref(io);
		if (!retval) {
			if (renamed) {
				tmp = g_strconcat(path, "~", NULL);		
				g_rename(tmp, path);
				g_free(tmp);
			} else
				g_unlink(path);
		}
	}
	g_free(path);

	return retval;
}

int main(int argc, char **argv)
{
	VDomain *domain = NULL;
	gint retval = 1;

	v_init(&argc, &argv, entries, NULL, NULL,
	       _("Sets up a virtual domain for its first use"));

	domain = v_domain_new(NULL, NULL, TRUE, NULL);
	if (!domain) {
		fprintf(stderr,
		        _("%s: Could not initialize domain handler.\n"),
		        g_get_prgname());
		goto out;
	}

	if (!setup_aliases(domain))
		goto out;

	if (v_exec_script("vsetup-pre")) {
		if (!quiet)
			fprintf(stderr, _("%s: Execution of 'vsetup-pre' "
			                  "failed.\n"), g_get_prgname());
		goto out;
	}

	if (!setup_usersdir(domain))
		goto out;

	if (!v_domain_commit_changes(domain, NULL)) {
		if (!quiet)
			fprintf(stderr, _("%s: Error commiting changes to "
			                  "the database.\n"), g_get_prgname());
		goto out;
	}

	if (!setup_qmail_handler(domain))
		goto out;

	if (v_exec_script("vsetup-post")) {
		if (!quiet)
			fprintf(stderr, _("%s: Execution of 'vsetup-pre' "
			                  "failed.\n"), g_get_prgname());
		goto out;
	}

	retval = 0;
out:
	if (domain)
		g_object_unref(domain);
	return retval;
}
